# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

PCSC_ABSENT = "This command requires the pyscard library that is not available on your system. Please consult https://docs.nitrokey.com/nitrokeys/nitrokey3/troubleshooting#pyscard-is-not-available for more information"



# Copyright 2019 SoloKeys Developers
# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import os
import platform
import struct
import sys
import time
from typing import Optional

if "linux" in platform.platform().lower():
    import fcntl

import click
from fido2.ctap import CtapError
from fido2.ctap1 import ApduError
from fido2.hid import CtapHidDevice

import pynitrokey.fido2 as nkfido2
from pynitrokey.cli.monitor import monitor
from pynitrokey.cli.program import program
from pynitrokey.cli.update import update
from pynitrokey.exceptions import NoSoloFoundError
from pynitrokey.helpers import local_critical, local_print, local_print_secret


@click.group()
def rng() -> None:
    """Access TRNG on device, see subcommands."""
    pass


@click.group()
def util() -> None:
    """Additional utilities, see subcommands."""
    pass


@click.command()
def list() -> None:
    """List all 'Nitrokey FIDO2' devices"""
    devs = nkfido2.find_all()
    local_print(":: 'Nitrokey FIDO2' keys")
    for c in devs:
        assert isinstance(c.dev, CtapHidDevice)
        descr = c.dev.descriptor

        if hasattr(descr, "product_name"):
            name = descr.product_name
        elif c.is_bootloader():
            name = "FIDO2 Bootloader device"
        else:
            name = "FIDO2 device"

        if hasattr(descr, "serial_number"):
            id_ = descr.serial_number
        else:
            assert isinstance(descr.path, str)
            id_ = descr.path

        local_print(f"{id_}: {name}")


@click.command()
@click.option("--count", default=8, help="How many bytes to generate (defaults to 8)")
@click.option(
    "-s",
    "--serial",
    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
)
def hexbytes(count: int, serial: Optional[str]) -> None:
    """Output COUNT number of random bytes, hex-encoded."""

    if not 0 <= count <= 255:
        local_critical(f"Number of bytes must be between 0 and 255, you passed {count}")
    local_print_secret(nkfido2.find(serial).get_rng(count).hex())


# @todo: not really useful like this? endless output only on request (--count ?)
@click.command()
@click.option(
    "-s",
    "--serial",
    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
)
def raw(serial: Optional[str]) -> None:
    """Output raw entropy endlessly."""
    p = nkfido2.find(serial)
    while True:
        r = p.get_rng(255)
        sys.stdout.buffer.write(r)


@click.command()
@click.option("--count", default=64, help="How many bytes to generate (defaults to 8)")
@click.option(
    "-s",
    "--serial",
    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
)
def feedkernel(count: int, serial: Optional[str]) -> None:
    """Feed random bytes to /dev/random."""

    if os.name != "posix":
        local_critical("This is a Linux-specific command!")

    if not 0 <= count <= 255:
        local_critical(f"Number of bytes must be between 0 and 255, you passed {count}")

    p = nkfido2.find(serial)

    RNDADDENTROPY = 0x40085203

    entropy_info_file = "/proc/sys/kernel/random/entropy_avail"
    print(f"entropy before: 0x{open(entropy_info_file).read().strip()}")

    r = p.get_rng(count)

    # man 4 random

    # RNDADDENTROPY
    #       Add some additional entropy to the input pool, incrementing the
    #       entropy count. This differs from writing to /dev/random or
    #       /dev/urandom, which only adds some data but does not increment the
    #       entropy count. The following structure is used:

    #           struct rand_pool_info {
    #               int    entropy_count;
    #               int    buf_size;
    #               __u32  buf[0];
    #           };

    #       Here entropy_count is the value added to (or subtracted from) the
    #       entropy count, and buf is the buffer of size buf_size which gets
    #       added to the entropy pool.

    # maximum 8, tend to be pessimistic
    entropy_bits_per_byte = 2
    t = struct.pack(f"ii{count}s", count * entropy_bits_per_byte, count, r)

    try:
        with open("/dev/random", mode="wb") as fh:
            fcntl.ioctl(fh, RNDADDENTROPY, t)

    except PermissionError as e:
        local_critical(
            "insufficient permissions to use `fnctl.ioctl` on '/dev/random'",
            "please run 'nitropy' with proper permissions",
            e,
        )

    local_print(f"entropy after:  0x{open(entropy_info_file).read().strip()}")


# @todo: also review, endless output only on request (--count ?)
@click.command()
@click.option(
    "-s",
    "--serial",
    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
)
@click.option("-b", "--blink", is_flag=True, help="Blink in the meantime")
def status(serial: Optional[str], blink: bool) -> None:
    """Print device's status"""
    p = nkfido2.find(serial)
    t0 = time.time()
    while True:
        if time.time() - t0 > 5 and blink:
            p.wink()
        r = p.get_status()
        for b in r:
            local_print("{:#02d} ".format(b), end="")
        local_print("")
        time.sleep(0.3)


@click.command()
@click.option(
    "-s",
    "--serial",
    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
)
def version(serial: Optional[str]) -> None:
    """Version of firmware on device."""

    try:
        res = nkfido2.find(serial).solo_version()
        major, minor, patch = res[:3]
        locked = ""
        # @todo:
        if len(res) > 3:
            if res[3]:  # type: ignore
                locked = "locked"
            else:
                locked = "unlocked"
        local_print(f"{major}.{minor}.{patch} {locked}")

    except NoSoloFoundError:
        local_critical(
            "No Nitrokey found.", "If you are on Linux, are your udev rules up to date?"
        )

    # unused ???
    except (NoSoloFoundError, ApduError):
        local_critical(
            "Firmware is out of date (key does not know the NITROKEY_VERSION command)."
        )


@click.command()
@click.option(
    "-s",
    "--serial",
    help="Serial number of Nitrokey to use. Prefix with 'device=' to provide device file, e.g. 'device=/dev/hidraw5'.",
)
def reboot(serial: Optional[str]) -> None:
    """Send reboot command to device (development command)"""
    local_print("Reboot", "Press key to confirm!")

    CTAP_REBOOT = 0x53
    dev = nkfido2.find(serial).dev
    try:
        assert isinstance(dev, CtapHidDevice)
        dev.call(CTAP_REBOOT ^ 0x80, b"")

    except OSError:
        local_print("...done")
    except CtapError as e:
        local_critical(f"...failed ({str(e)})")


rng.add_command(hexbytes)
rng.add_command(raw)
rng.add_command(feedkernel)

util.add_command(program)
util.add_command(monitor)


def add_commands(fido2: click.Group) -> None:

    fido2.add_command(list)
    # @fixme: this one exists twice, once here, once in "util program aux"
    fido2.add_command(reboot)
    fido2.add_command(rng)
    fido2.add_command(status)
    fido2.add_command(util)
    fido2.add_command(version)
    fido2.add_command(update)



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import logging
import os
from typing import Any, Callable, Optional, Sequence, Union

import smartcard
from cryptography.hazmat.decrepit.ciphers.algorithms import TripleDES
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from smartcard.CardConnection import CardConnection
from smartcard.Exceptions import NoCardException

from pynitrokey.helpers import local_critical
from pynitrokey.start.gnuk_token import iso7816_compose
from pynitrokey.tlv import Tlv

LogFn = Callable[[str], Any]


def find_by_id(tag: int, data: Sequence[tuple[int, bytes]]) -> Optional[bytes]:
    for t, b in data:
        if t == tag:
            return b
    return None


# size is in bytes
def prepare_for_pkcs1v15_sign(data: bytes, key_size_bytes: int) -> bytes:
    digest = hashes.Hash(hashes.SHA256())
    digest.update(data)
    hashed = digest.finalize()

    # ASN.1 DigestInfo prefix for SHA-256
    prefix = bytearray.fromhex("3031300d060960864801650304020105000420")

    # PKCS#1 v1.5 block:
    #  0x00 0x01 FF FF ... FF 0x00 prefix hash
    padding_len = key_size_bytes - 3 - len(prefix) - len(hashed)
    padding = b"\x00\x01" + (b"\xff" * padding_len) + b"\x00"

    total = padding + prefix + hashed
    assert len(total) == key_size_bytes
    return total


class StatusError(Exception):
    def __init__(self, value: int):
        self.value = value

    def __str__(self) -> str:
        return f"{hex(self.value)}"


class PivApp:
    log: logging.Logger
    logfn: LogFn
    connection: CardConnection

    def __init__(self, logfn: Optional[LogFn] = None):
        self.log = logging.getLogger("pivapp")
        readers = smartcard.System.readers()
        chosen_connection: Optional[CardConnection] = None
        for r in readers:
            connection = r.createConnection()
            try:
                connection.connect()
            except NoCardException:
                continue

            expected_atr = list(bytes.fromhex("3b8f01805d4e6974726f6b657900000000006a"))
            if not expected_atr == connection.getATR():
                continue
            print(r)

            select = [
                0x00,
                0xA4,
                0x04,
                0x00,
                0x0C,
                0xA0,
                0x00,
                0x00,
                0x03,
                0x08,
                0x00,
                0x00,
                0x10,
                0x00,
                0x01,
                0x00,
                0x00,
                0x00,
            ]
            data, sw1, sw2 = connection.transmit(select)
            if sw1 != 0x90 or sw2 != 0x00:
                continue
            chosen_connection = connection

        if not chosen_connection:
            raise NoCardException("No PIV card found", -1)

        self.connection = chosen_connection

        if logfn is not None:
            self.logfn = logfn
        else:
            self.logfn = self.log.info

    def send_receive(
        self,
        ins: int,
        p1: int,
        p2: int,
        data: bytes = b"",
    ) -> bytes:
        bytes_data = iso7816_compose(ins, p1, p2, data)
        return self._send_receive_inner(bytes_data, log_info=f"{ins}")

    def _send_receive_inner(self, data: bytes, log_info: str = "") -> bytes:
        self.logfn(
            f"Sending {log_info if log_info else ''} {data.hex() if data else data!r}"
        )

        try:
            result_list, sw1, sw2 = self.connection.transmit(list(data))
        except Exception as e:
            self.logfn(f"Got exception: {e}")
            raise

        result = bytes(result_list)
        status_bytes = bytes([sw1, sw2])
        self.logfn(f"Received [{status_bytes.hex()}] {result.hex()}")

        log_multipacket = False
        data_final = bytes(result)
        MORE_DATA_STATUS_BYTE = 0x61
        while status_bytes[0] == MORE_DATA_STATUS_BYTE:
            if log_multipacket:
                self.logfn(
                    f"Got RemainingData status: [{status_bytes.hex()}] {result.hex() if result else result!r}"
                )
            log_multipacket = True
            ins = 0xC0
            p1 = 0
            p2 = 0
            le = sw2 if sw2 != 0 else 0xFF
            bytes_data = iso7816_compose(ins, p1, p2, le=le)
            try:
                result_list, sw1, sw2 = self.connection.transmit(list(bytes_data))
            except Exception as e:
                self.logfn(f"Got exception: {e}")
                raise
            # Data order is different here than in APDU - SW is first, then the data if any
            result = bytes(result_list)
            status_bytes = bytes([sw1, sw2])
            self.logfn(f"Received [{status_bytes.hex()}] {bytes(result).hex()}")
            if status_bytes[0] in [0x90, MORE_DATA_STATUS_BYTE]:
                data_final += bytes(result)

        if status_bytes != b"\x90\x00" and status_bytes[0] != MORE_DATA_STATUS_BYTE:
            raise StatusError(int.from_bytes(status_bytes, byteorder="big"))

        if log_multipacket:
            self.logfn(
                f"Received final data: [{status_bytes.hex()}] {data_final.hex() if data_final else data_final!r}"
            )

        if data_final:
            try:
                self.logfn(f"Decoded received: {data_final.hex()}")
            except Exception:
                pass

        return data_final

    def authenticate_admin(self, admin_key: bytes) -> None:

        if len(admin_key) == 24:
            algorithm: Union[TripleDES, algorithms.AES128, algorithms.AES256] = (
                TripleDES(admin_key)
            )
            # algo = "tdes"
            algo_byte = 0x03
            expected_len = 8
        elif len(admin_key) == 16:
            algorithm = algorithms.AES128(admin_key)
            # algo = "aes128"
            algo_byte = 0x08
            expected_len = 16
        elif len(admin_key) == 32:
            algorithm = algorithms.AES256(admin_key)
            # algo = "aes256"
            algo_byte = 0x0C
            expected_len = 16
        else:
            local_critical(
                "Unsupported key length",
                support_hint=False,
            )

        challenge_body = Tlv.build([(0x7C, Tlv.build([(0x80, b"")]))])
        challenge_response = self.send_receive(0x87, algo_byte, 0x9B, challenge_body)
        general_auth_data = find_by_id(0x7C, Tlv.parse(challenge_response))
        if general_auth_data is None:
            local_critical("Failed to get response to GENERAL AUTHENTICATE")
            return

        challenge = find_by_id(
            0x80,
            Tlv.parse(general_auth_data),
        )

        if challenge is None:
            local_critical("Failed to get authentication challenge from the device")
            return

        # challenge = decoded.first_by_id(0x7C).data.first_by_id(0x80).data
        if len(challenge) != expected_len:
            local_critical("Got unexpected authentication challenge length")

        our_challenge = os.urandom(expected_len)
        cipher = Cipher(algorithm, mode=modes.ECB())
        decryptor = cipher.decryptor()
        response = decryptor.update(challenge) + decryptor.finalize()
        decryptor = cipher.decryptor()
        our_challenge_encrypted = decryptor.update(our_challenge) + decryptor.finalize()
        response_body = Tlv.build(
            [(0x7C, Tlv.build([(0x80, response), (0x81, our_challenge_encrypted)]))]
        )

        final_response = self.send_receive(0x87, algo_byte, 0x9B, response_body)
        general_auth_data = find_by_id(0x7C, Tlv.parse(final_response))
        if general_auth_data is None:
            local_critical("Failed to get response to GENERAL AUTHENTICATE")
            return

        decoded_challenge = find_by_id(
            0x82,
            Tlv.parse(general_auth_data),
        )

        if decoded_challenge != our_challenge:
            local_critical(
                "Failed to authenticate with administrator key", support_hint=False
            )

    def set_admin_key(self, new_key: bytes) -> None:
        if len(new_key) == 24:
            # algo = "tdes"
            algo_byte = 0x03
        elif len(new_key) == 16:
            # algo = "aes128"
            algo_byte = 0x08
        elif len(new_key) == 32:
            # algo = "aes256"
            algo_byte = 0x0C
        else:
            local_critical(
                "Unsupported key length",
                support_hint=False,
            )
        data = bytes([algo_byte, 0x9B, len(new_key)]) + new_key
        self.send_receive(0xFF, 0xFF, 0xFE, data)

    def encode_pin(self, pin: str) -> bytes:
        body = pin.encode("utf-8")
        if len(body) > 8:
            local_critical("PIN can only be up to 8 bytes long", support_hint=False)

        body += bytes([0xFF for i in range(8 - len(body))])
        return body

    def login(self, pin: str) -> None:
        body = self.encode_pin(pin)
        self.send_receive(0x20, 0x00, 0x80, body)

    def change_pin(self, old_pin: str, new_pin: str) -> None:
        body = self.encode_pin(old_pin) + self.encode_pin(new_pin)
        self.send_receive(0x24, 0, 0x80, body)

    def change_puk(self, old_puk: str, new_puk: str) -> None:
        old_puk_bytes = old_puk.encode("utf-8")
        new_puk_bytes = new_puk.encode("utf-8")
        if len(old_puk_bytes) != 8 or len(new_puk) != 8:
            local_critical("PUK must be 8 bytes long", support_hint=False)
        body = old_puk_bytes + new_puk_bytes
        self.send_receive(0x24, 0, 0x81, body)

    def reset_retry_counter(self, puk: str, new_pin: str) -> None:
        puk_bytes = puk.encode("utf-8")

        if len(puk_bytes) != 8:
            local_critical("PUK must be 8 bytes long", support_hint=False)

        body = puk_bytes + self.encode_pin(new_pin)
        self.send_receive(0x2C, 0, 0x80, body)

    def factory_reset(self) -> None:
        self.send_receive(0xFB, 0, 0)

    def sign_p256(self, data: bytes, key: int) -> bytes:
        digest = hashes.Hash(hashes.SHA256())
        digest.update(data)
        payload = digest.finalize()
        return self.raw_sign(payload, key, 0x11)

    def sign_rsa(self, data: bytes, key: int, bits: int) -> bytes:
        key_size_bytes = bits // 8
        payload = prepare_for_pkcs1v15_sign(data, key_size_bytes)
        algo_map = {2048: 0x07, 3072: 0x05, 4096: 0x16}
        algo_id = algo_map[bits]
        return self.raw_sign(payload, key, algo_id)

    def raw_sign(self, payload: bytes, key: int, algo: int) -> bytes:
        body = Tlv.build([(0x7C, Tlv.build([(0x81, payload), (0x82, b"")]))])
        result = self.send_receive(0x87, algo, key, body)
        general_auth_data = find_by_id(0x7C, Tlv.parse(result))
        if general_auth_data is None:
            local_critical("Failed to get response to GENERAL AUTHENTICATE")
            return bytes()

        signature = find_by_id(
            0x82,
            Tlv.parse(general_auth_data),
        )

        if signature is None:
            local_critical("Failed to get signature from device")
            # Satisfy the type checker.
            # local_critical raises always raises an error
            return b""

        return signature

    def init(self) -> bytes:
        # Template for card capabilities with nothing but a random ID
        template_begin = bytearray.fromhex("f015a000000116")
        template_end = bytearray.fromhex(
            "f10121f20121f300f40100f50110f600f700fa00fb00fc00fd00fe00"
        )
        card_id = os.urandom(16)
        cardcaps = template_begin + card_id + template_end
        cardcaps_body = Tlv.build(
            [(0x5C, bytes(bytearray.fromhex("5fc107"))), (0x53, bytes(cardcaps))]
        )
        self.send_receive(0xDB, 0x3F, 0xFF, cardcaps_body)

        pinfo_body = Tlv.build(
            [
                (0x5C, bytes(bytearray.fromhex("5FC109"))),
                (
                    0x53,
                    Tlv.build(
                        [
                            (0x01, "Nitrokey PIV user".encode("ascii")),
                            # TODO: use representation of real serial number of card (currently static value)
                            # Base 10 representation of
                            # https://github.com/Nitrokey/piv-authenticator/blob/2c948a966f3e410e9a4cee3c351ca20b956383e0/src/lib.rs#L197
                            (0x05, "5437251".encode("ascii")),
                        ]
                    ),
                ),
            ]
        )
        self.send_receive(0xDB, 0x3F, 0xFF, pinfo_body)
        return card_id

    def serial(self) -> int:
        response = self.send_receive(0x01, 0x00, 0x00)
        return int.from_bytes(response, byteorder="big")

    def reader(self) -> str:
        reader: str = self.connection.getReader()  # type: ignore
        return reader

    def guid(self) -> bytes:
        payload = Tlv.build([(0x5C, bytes(bytearray.fromhex("5FC102")))])
        chuid = self.send_receive(0xCB, 0x3F, 0xFF, payload)

        chuid_tmp = find_by_id(0x53, Tlv.parse(chuid))
        if chuid_tmp is None:
            local_critical("Failed to get chuid from device")
            return b""

        chuid_data = find_by_id(0x34, Tlv.parse(chuid_tmp))
        if chuid_data is None:
            local_critical("Failed to get chuid from device")
            # Satisfy the type checker.
            # local_critical raises always raises an error
            return b""

        return chuid_data

    def cert(self, container_id: bytes) -> Optional[bytes]:
        payload = Tlv.build([(0x5C, container_id)])
        try:
            cert = self.send_receive(0xCB, 0x3F, 0xFF, payload)
            parsed = Tlv.parse(cert)
            if len(parsed) != 1:
                local_critical("Bad number of elements", support_hint=False)

            tag, value = parsed[0]
            if tag != 0x53:
                local_critical("Bad tag", support_hint=False)

            parsed = Tlv.parse(value)
            if len(parsed) < 1:
                local_critical("Bad number of sub-elements", support_hint=False)

            tag, value = parsed[0]
            if tag != 0x70:
                local_critical("Bad tag", support_hint=False)

            return value

        except StatusError as e:
            if e.value == 0x6A82:
                return None
            else:
                raise ValueError(f"{hex(e.value)}, Received error")



# Copyright 2019 SoloKeys Developers
# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT


class SoloExtension:
    version = 0x14
    rng = 0x15


class SoloBootloader:
    write = 0x40
    done = 0x41
    check = 0x42
    erase = 0x43
    version = 0x44
    reboot = 0x45
    st_dfu = 0x46
    disable = 0x47
    boot_pubkey = 0x48

    HIDCommandBoot = 0x50
    HIDCommandEnterBoot = 0x51
    HIDCommandEnterSTBoot = 0x52
    HIDCommandRNG = 0x60
    HIDCommandProbe = 0x70
    HIDCommandStatus = 0x71

    TAG = b"\x8c\x27\x90\xf6"



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

from typing import Optional, Sequence, Tuple


def build_one(tag: int, data: bytes) -> bytes:
    data_len = len(data)
    out = bytearray()
    out += tag.to_bytes((tag.bit_length() + 7) // 8, byteorder="big")
    if data_len <= 0x7F:
        out.append(data_len)
    elif data_len <= 0xFF:
        out.append(0x81)
        out.append(data_len)
    else:
        assert data_len <= 0xFFFF
        out.append(0x82)
        out += data_len.to_bytes((data_len.bit_length() + 7) // 8, byteorder="big")

    return out + data


def take_tag(data: bytes) -> Tuple[int, bytes]:
    data_len = len(data)
    if data_len == 0:
        raise ValueError("Failed to parse TLV data: empty data when parsing tag")

    b1 = data[0]
    if (b1 & 0x1F) == 0x1F:
        if data_len < 2:
            raise ValueError("Failed to parse TLV data: partial tag")
        b2 = data[1]
        return (int.from_bytes([b1, b2], byteorder="big"), data[2:])
    else:
        return (int.from_bytes([0, b1], byteorder="big"), data[1:])


def take_len(data: bytes) -> Tuple[int, bytes]:
    data_len = len(data)
    if data_len == 0:
        raise ValueError("Failed to parse TLV data: empty data when parsing len")

    l1 = data[0]
    if l1 <= 0x7F:
        return (l1, data[1:])
    elif l1 == 0x81:
        if data_len < 2:
            raise ValueError("Failed to parse TLV data: partial len")
        return (data[1], data[2:])
    elif l1 == 0x82:
        if data_len < 3:
            raise ValueError("Failed to parse TLV data: partial len")
        l2 = data[1]
        l3 = data[2]
        return (int.from_bytes([l2, l3], byteorder="big"), data[3:])
    else:
        raise ValueError("Failed to parse TLV data: invalid len")


def take_do(data: bytes) -> Tuple[int, bytes, bytes]:
    tag, rem = take_tag(data)
    len, rem = take_len(rem)

    return tag, rem[:len], rem[len:]


class Tlv:
    @staticmethod
    def build(input: Sequence[Tuple[int, bytes]]) -> bytes:
        out = bytearray()
        for tag, data in input:
            out += build_one(tag, data)
        return out

    @staticmethod
    def parse(data: bytes) -> Sequence[Tuple[int, bytes]]:
        res = []
        current = data
        while len(current) != 0:
            tag, value, rem = take_do(current)
            res.append((tag, value))
            current = rem
        return res



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

from typing import Optional

import click
from nitrokey.nkpk import NKPK, NKPKBootloader
from nitrokey.trussed import Model

from pynitrokey.cli.trussed.test import TestCase

from . import trussed


class Context(trussed.Context[NKPKBootloader, NKPK]):
    def __init__(self, path: Optional[str]) -> None:
        super().__init__(
            path,
            NKPKBootloader,
            NKPK,
            Model.NKPK,
        )

    @property
    def test_cases(self) -> list[TestCase]:
        from pynitrokey.cli.trussed import tests

        return [
            tests.test_uuid_query,
            tests.test_firmware_version_query,
            tests.test_nkpk_device_status,
            tests.test_bootloader_configuration,
            tests.test_firmware_mode,
            tests.test_fido2,
        ]


@click.group()
@click.option("-p", "--path", "path", help="The path of the Nitrokey 3 device")
@click.pass_context
def nkpk(ctx: click.Context, path: Optional[str]) -> None:
    """Interact with Nitrokey Passkey devices, see subcommands."""
    ctx.obj = Context(path)
    trussed.prepare_group()


# shared Trussed commands
trussed.add_commands(nkpk, has_app_reset=False)


def _list() -> None:
    trussed._list(Context(None))



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import sys
import time
from typing import List, Optional, Union

from fido2.hid import CtapHidDevice

from pynitrokey.exceptions import NoSoloFoundError
from pynitrokey.fido2.client import NKFido2Client, list_ctaphid_devices


def find(
    solo_serial: Optional[str] = None,
    retries: int = 5,
    raw_device: Optional[CtapHidDevice] = None,
) -> NKFido2Client:
    p = NKFido2Client()

    # This... is not the right way to do it yet
    p.use_u2f()

    for i in range(retries):
        try:
            p.find_device(dev=raw_device, solo_serial=solo_serial)
            return p
        except RuntimeError:
            time.sleep(0.2)

    print(
        "Warning: This command only works with the Nitrokey FIDO2, not with "
        "other FIDO2 devices!",
        file=sys.stderr,
    )
    raise NoSoloFoundError("no Nitrokey FIDO2 found")


def find_all() -> List[NKFido2Client]:
    return [find(raw_device=device) for device in list_ctaphid_devices()]


def device_path_to_str(path: Union[bytes, str]) -> str:
    """
    Converts a device path as returned by the fido2 library to a string.

    Typically, the path already is a string.  Only on Windows, a bytes object
    using an ANSI encoding is used instead.  We use the ISO 8859-1 encoding to
    decode the string which should work for all systems.
    """
    if isinstance(path, bytes):
        return path.decode("iso-8859-1", errors="ignore")
    else:
        return path



# Copyright 2019 SoloKeys Developers
# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

"""Python Library for Nitrokey devices."""

__all__ = ["client", "commands", "dfu", "enums", "exceptions", "helpers", "operations"]



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import logging
import platform
import sys
from dataclasses import dataclass
from enum import Enum, auto, unique
from types import TracebackType
from typing import Callable, Iterable, Optional, Sequence, Tuple, Type, Union

from nitrokey.trussed import TrussedBase, Version

from pynitrokey.cli.exceptions import CliException
from pynitrokey.fido2 import device_path_to_str
from pynitrokey.helpers import local_print

logger = logging.getLogger(__name__)

DEFAULT_EXCLUDES = ["bootloader", "provisioner"]


ExcInfo = Tuple[Type[BaseException], BaseException, TracebackType]


class TestContext:
    def __init__(self, pin: Optional[str]) -> None:
        self.pin = pin
        self.firmware_version: Optional[Version] = None


@unique
class TestStatus(Enum):
    SKIPPED = auto()
    SUCCESS = auto()
    FAILURE = auto()


class TestResult:
    def __init__(
        self,
        status: TestStatus,
        data: Optional[str] = None,
        exc_info: Union[ExcInfo, Tuple[None, None, None]] = (None, None, None),
    ) -> None:
        self.status = status
        self.data = data
        self.exc_info = exc_info


TestCaseFn = Callable[[TestContext, TrussedBase], TestResult]


class TestCase:
    def __init__(self, name: str, description: str, fn: TestCaseFn) -> None:
        self.name = name
        self.description = description
        self.fn = fn


def test_case(name: str, description: str) -> Callable[[TestCaseFn], TestCase]:
    def decorator(func: TestCaseFn) -> TestCase:
        return TestCase(name, description, func)

    return decorator


def filter_test_cases(
    test_cases: Sequence[TestCase], names: Iterable[str]
) -> Iterable[TestCase]:
    for test_case in test_cases:
        if test_case.name in names:
            yield test_case


@dataclass
class TestSelector:
    only: Iterable[str] = ()
    all: bool = False
    include: Iterable[str] = ()
    exclude: Iterable[str] = ()

    def select(self, test_cases: Sequence[TestCase]) -> list[TestCase]:
        if self.only:
            return list(filter_test_cases(test_cases, self.only))

        selected = []
        for test_case in test_cases:
            if test_case.name in self.include:
                selected.append(test_case)
            elif test_case.name not in self.exclude:
                if self.all or test_case.name not in DEFAULT_EXCLUDES:
                    selected.append(test_case)
        return selected


def log_devices() -> None:
    from fido2.hid import CtapHidDevice

    ctap_devices = [device for device in CtapHidDevice.list_devices()]
    logger.info(f"Found {len(ctap_devices)} CTAPHID devices:")
    for device in ctap_devices:
        descriptor = device.descriptor
        path = device_path_to_str(descriptor.path)
        logger.info(f"- {path} ({descriptor.vid:x}:{descriptor.pid:x})")


def log_system() -> None:
    logger.info(f"platform: {platform.platform()}")
    logger.info(f"uname: {platform.uname()}")


def list_tests(
    selector: TestSelector,
    test_cases: Sequence[TestCase],
) -> None:
    test_cases = selector.select(test_cases)
    print(f"{len(test_cases)} test case(s) selected")
    for test_case in test_cases:
        print(f"- {test_case.name}: {test_case.description}")


def run_tests(
    ctx: TestContext,
    device: TrussedBase,
    selector: TestSelector,
    test_cases: Sequence[TestCase],
) -> bool:
    test_cases = selector.select(test_cases)
    if not test_cases:
        raise CliException("No test cases selected", support_hint=False)

    results = []

    local_print("")
    local_print(f"Running tests for {device.name} at {device.path}")
    local_print("")

    n = len(test_cases)
    idx_len = len(str(n))
    name_len = max([len(test_case.name) for test_case in test_cases]) + 2
    description_len = max([len(test_case.description) for test_case in test_cases]) + 2
    status_len = max([len(status.name) for status in TestStatus]) + 2

    for i, test_case in enumerate(test_cases):
        try:
            result = test_case.fn(ctx, device)
        except Exception:
            result = TestResult(TestStatus.FAILURE, exc_info=sys.exc_info())
        results.append(result)

        idx = str(i + 1).rjust(idx_len)
        name = test_case.name.ljust(name_len)
        description = test_case.description.ljust(description_len)
        status = result.status.name.ljust(status_len)
        msg = ""
        if result.data:
            msg = str(result.data)
        elif result.exc_info[1]:
            logger.error(
                f"An exception occured during the execution of the test {test_case.name}:",
                exc_info=result.exc_info,
            )
            msg = str(result.exc_info[1])

        local_print(f"[{idx}/{n}]\t{name}\t{description}\t{status}\t{msg}")

    success = len([result for result in results if result.status == TestStatus.SUCCESS])
    skipped = len([result for result in results if result.status == TestStatus.SKIPPED])
    failed = len([result for result in results if result.status == TestStatus.FAILURE])
    local_print("")
    local_print(f"{n} tests, {success} successful, {skipped} skipped, {failed} failed")

    return all([result.status != TestStatus.FAILURE for result in results])



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import logging
from collections.abc import Set
from contextlib import contextmanager
from typing import Any, Callable, Iterator, List, Optional

from click import Abort
from nitrokey.trussed import Model, TrussedBootloader, TrussedDevice, Version
from nitrokey.trussed.admin_app import Status
from nitrokey.trussed.updates import DeviceHandler, Updater, UpdateUi, Warning

from pynitrokey.cli.exceptions import CliException
from pynitrokey.cli.trussed import Bootloader, Context, Device
from pynitrokey.helpers import DownloadProgressBar, ProgressBar, confirm, local_print

logger = logging.getLogger(__name__)


class UpdateCli(UpdateUi):
    def __init__(self, confirm_continue: bool = False) -> None:
        self._version_printed = False
        self._confirm_continue = confirm_continue

    def error(self, *msgs: Any) -> Exception:
        return CliException(*msgs)

    def abort(self, *msgs: Any) -> Exception:
        return CliException(*msgs, support_hint=False)

    def raise_warning(self, warning: Warning) -> Exception:
        return self.abort(
            f"{warning.message}\nTo ignore this warning and install the update at your own risk,"
            f" set the --ignore-warning {warning.value} option."
        )

    def show_warning(self, warning: Warning) -> None:
        logger.warning(f"Ignoring warning {warning.value}")
        local_print(f"Warning: {warning.message}")
        local_print(
            f"Note: The update will continue as --ignore-warning {warning.value} has been set."
        )

    def abort_downgrade(self, current: Version, image: Version) -> Exception:
        self._print_firmware_versions(current, image)
        return self.abort(
            "The firmware image is older than the firmware on the device."
        )

    def abort_pynitrokey_version(
        self, current: Version, required: Version
    ) -> Exception:
        return self.abort(
            f"This update requires pynitrokey version {required} (current: {current}). "
            "Please update pynitrokey to install the update."
        )

    def confirm_download(self, current: Optional[Version], new: Version) -> None:
        if self._confirm_continue:
            return

        confirm(
            f"Do you want to download the firmware version {new}?",
            default=True,
            abort=True,
        )

    def confirm_pynitrokey_version(self, current: Version, required: Version) -> None:
        local_print(
            f"This update requires pynitrokey version {required} (current: {current})."
        )
        local_print("Using an outdated pynitrokey version is strongly discouraged.")
        if not confirm(
            "Do you want to continue with an outdated pynitrokey version at your own risk?"
        ):
            logger.info("Update cancelled by user")
            raise Abort()

    def confirm_update(self, current: Optional[Version], new: Version) -> None:
        self._print_firmware_versions(current, new)
        local_print("")
        local_print(
            "Please do not remove the device or insert any other Nitrokey devices "
            "during the update. Doing so may damage the device."
        )

        if self._confirm_continue:
            return

        if not confirm("Do you want to perform the firmware update now?"):
            logger.info("Update cancelled by user")
            raise Abort()

    def confirm_update_same_version(self, version: Version) -> None:
        self._print_firmware_versions(version, version)
        if not confirm(
            "The version of the firmware image is the same as on the device.  Do you want "
            "to continue anyway?"
        ):
            raise Abort()

    def confirm_extra_information(self, txt: List[str]) -> None:
        if txt:
            local_print("\n".join(txt))
            if not confirm("Have you read these information? Do you want to continue?"):
                raise Abort()

    def pre_bootloader_hint(self) -> None:
        pass

    def request_bootloader_confirmation(self) -> None:
        local_print("")
        local_print(
            "Please press the touch button to reboot the device into bootloader mode ..."
        )
        local_print("")

    @contextmanager
    def download_progress_bar(self, desc: str) -> Iterator[Callable[[int, int], None]]:
        with DownloadProgressBar(desc) as bar:
            yield bar.update

    @contextmanager
    def update_progress_bar(self) -> Iterator[Callable[[int, int], None]]:
        with ProgressBar(
            desc="Perform firmware update", unit="B", unit_scale=True
        ) as bar:
            yield bar.update_sum

    @contextmanager
    def finalization_progress_bar(self) -> Iterator[Callable[[int, int], None]]:
        with ProgressBar(desc="Finalize upgrade", unit="%", unit_scale=False) as bar:
            yield bar.update_sum

    def _print_firmware_versions(
        self, current: Optional[Version], new: Optional[Version]
    ) -> None:
        if not self._version_printed:
            current_str = str(current) if current else "[unknown]"
            local_print(f"Current firmware version:  {current_str}")
            local_print(f"Updated firmware version:  {new}")
            self._version_printed = True


class ContextDeviceHandler(DeviceHandler):
    def __init__(self, ctx: Context[Bootloader, Device]) -> None:
        self.ctx = ctx

    def await_bootloader(self, model: Model) -> TrussedBootloader:
        assert model == self.ctx.model
        return self.ctx.await_bootloader()

    def await_device(
        self,
        model: Model,
        wait_retries: Optional[int],
        callback: Optional[Callable[[int, int], None]],
    ) -> TrussedDevice:
        assert model == self.ctx.model
        return self.ctx.await_device(wait_retries, callback)


def update(
    ctx: Context[Bootloader, Device],
    image: Optional[str],
    version: Optional[str],
    ignore_pynitrokey_version: bool,
    ignore_warnings: Set[Warning],
    confirm_continue: bool,
) -> tuple[Version, Status]:
    with ctx.connect() as device:
        updater = Updater(
            ui=UpdateCli(confirm_continue),
            device_handler=ContextDeviceHandler(ctx),
            ignore_warnings=ignore_warnings,
        )
        return updater.update(device, image, version, ignore_pynitrokey_version)



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

from typing import Optional

import click
from nitrokey.nk3 import NK3, NK3Bootloader
from nitrokey.trussed import Model

from pynitrokey.cli import trussed
from pynitrokey.cli.trussed.test import TestCase


class Context(trussed.Context[NK3Bootloader, NK3]):
    def __init__(self, path: Optional[str]) -> None:
        super().__init__(path, NK3Bootloader, NK3, Model.NK3)  # type: ignore[type-abstract]

    @property
    def test_cases(self) -> list[TestCase]:
        from pynitrokey.cli.trussed import tests

        return [
            tests.test_uuid_query,
            tests.test_firmware_version_query,
            tests.test_nk3_device_status,
            tests.test_bootloader_configuration,
            tests.test_firmware_mode,
            tests.test_se050,
            tests.test_fido2,
        ]


@click.group()
@click.option("-p", "--path", "path", help="The path of the Nitrokey 3 device")
@click.pass_context
def nk3(ctx: click.Context, path: Optional[str]) -> None:
    """Interact with Nitrokey 3 devices, see subcommands."""
    ctx.obj = Context(path)
    trussed.prepare_group()


# shared Trussed commands
trussed.add_commands(nk3)


def _list() -> None:
    trussed._list(Context(None))


@nk3.command()
@click.pass_obj
def wink(ctx: Context) -> None:
    """Send wink command to the device (blinks LED a few times)."""
    with ctx.connect_device() as device:
        device.wink()


# This import has to be added here to avoid circular dependency
# Import "secrets" subcommand from the secrets module
from . import piv  # noqa: F401,E402
from . import secrets  # noqa: F401,E402



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

from typing import Any

from pynitrokey.helpers import local_critical


class CliException(Exception):
    def __init__(
        self,
        *messages: Any,
        support_hint: bool = True,
        ret_code: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__("\n".join([str(message) for message in messages]))

        self.messages = messages
        self.support_hint = support_hint
        self.ret_code = ret_code
        self.kwargs = kwargs

    def show(self) -> None:
        local_critical(
            *self.messages,
            support_hint=self.support_hint,
            ret_code=self.ret_code,
            **self.kwargs,
        )



#! /usr/bin/env python3

# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import logging
import subprocess
import threading
import time
from sys import stderr
from typing import Iterable, List, Optional


class ThreadLog(threading.Thread):
    _dmesg_skip_strings = []  # type: ignore

    _write_to_log = False

    def __init__(self, logger: logging.Logger, command: str):
        threading.Thread.__init__(self)
        self.finished = False
        self.logger = logger
        self.command = command
        self.daemon = True
        self.start()
        self.process: Optional[subprocess.Popen[bytes]] = None

    def run(self):
        self.execute(self.command.split())

    @staticmethod
    def _contains(value: str, strings: Iterable[str]) -> bool:
        for s in strings:
            if s in value:
                return True
        return False

    def execute(self, command: List[str]):
        self.process = subprocess.Popen(
            command, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )

        # Poll process for new output until finished
        for line in iter(self.process.stdout.readline, ""):  # type: ignore
            if self.finished:
                break
            if not line or self.finished or not self._write_to_log:
                continue
            if self._contains(line, self._dmesg_skip_strings):
                continue
            self.logger.debug(line.strip())

        self.process.wait()
        self.logger.debug("Finished")

    def start_logging(self) -> None:
        self._write_to_log = True

    def __enter__(self):
        time.sleep(1)
        self.start_logging()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.debug("Finishing")
        self.finished = True
        self.process.kill()
        self.process.wait()
        self.join(10)


def test_run() -> None:
    FORMAT = "%(relativeCreated)05d [%(process)x] - %(levelname)s - %(name)s - %(message)s [%(filename)s:%(lineno)d]"
    logging.basicConfig(format=FORMAT, stream=stderr, level=logging.DEBUG)
    logger = logging.getLogger("threadlog")

    try:
        t = ThreadLog(logger, "dmesg -w")
        t.start_logging()
        time.sleep(10)
        t.finished = True
    except (KeyboardInterrupt, SystemExit) as k:
        return


if __name__ == "__main__":
    test_run()



# Copyright 2019 SoloKeys Developers
# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import base64
import json
import struct
import sys
import tempfile
import time
from typing import Optional, Tuple, Union

from fido2.ctap import CtapError
from fido2.ctap1 import Ctap1
from fido2.ctap2.base import Ctap2
from fido2.hid import CTAPHID, CtapHidDevice, open_device
from intelhex import IntelHex

import pynitrokey.exceptions
from pynitrokey import helpers
from pynitrokey.fido2.commands import SoloBootloader, SoloExtension
from pynitrokey.helpers import local_critical


def list_ctaphid_devices() -> list[CtapHidDevice]:
    hid_devices = list(CtapHidDevice.list_devices())
    return [
        d
        for d in hid_devices
        if (d.descriptor.vid, d.descriptor.pid)
        in [
            (0x0483, 0xA2CA),  #
            (0x20A0, 0x42B3),  # ...
            (0x20A0, 0x42B1),  # NK FIDO2
        ]
    ]


class NKFido2Client:
    def __init__(self) -> None:
        self.origin = "https://example.org"
        self.host = "example.org"
        self.user_id = b"they"
        self.exchange = self.exchange_hid
        self.do_reboot = True

    def use_u2f(self) -> None:
        self.exchange = self.exchange_u2f

    def use_hid(self) -> None:
        self.exchange = self.exchange_hid

    def reboot(self) -> None:
        """option to reboot after programming"""
        try:
            self.exchange(SoloBootloader.reboot)
        except OSError:
            pass

    def find_device(
        self,
        dev: Optional[CtapHidDevice] = None,
        solo_serial: Optional[str] = None,
    ) -> CtapHidDevice:

        devices = []
        found_dev: Optional[CtapHidDevice] = None

        if dev is None:
            if solo_serial is not None:
                if solo_serial.startswith("device="):
                    solo_serial = solo_serial.split("=")[1]
                    found_dev = open_device(solo_serial)
                else:
                    devices = list_ctaphid_devices()
                    devices = [
                        d for d in devices if d.descriptor.serial_number == solo_serial
                    ]
            else:
                devices = list_ctaphid_devices()
            if len(devices) > 1:
                raise pynitrokey.exceptions.NonUniqueDeviceError
            if len(devices) > 0:
                found_dev = devices[0]
        if dev is None and found_dev is None:
            raise RuntimeError("No FIDO device found")

        self.dev = found_dev if found_dev else dev
        assert isinstance(self.dev, CtapHidDevice)
        self.ctap1 = Ctap1(self.dev)

        try:
            self.ctap2: Optional[Ctap2] = Ctap2(self.dev)
        except CtapError:
            self.ctap2 = None

        if self.exchange == self.exchange_hid:
            self.send_data_hid(CTAPHID.INIT, b"\x11\x11\x11\x11\x11\x11\x11\x11")

        return self.dev

    @staticmethod
    def format_request(cmd: int, addr: int = 0, data: bytes = b"A" * 16) -> bytes:
        # not sure why this is here?
        # arr = b"\x00" * 9
        packed_addr = struct.pack("<L", addr)
        packed_cmd = struct.pack("B", cmd)
        length = struct.pack(">H", len(data))

        return packed_cmd + packed_addr[:3] + SoloBootloader.TAG + length + data

    def send_data_hid(self, cmd: int, data: bytes = b"A" * 16) -> bytes:
        if not isinstance(data, bytes):
            data = struct.pack("%dB" % len(data), *[ord(x) for x in data])
        with helpers.Timeout(1.0) as event:
            assert isinstance(self.dev, CtapHidDevice)
            return self.dev.call(cmd, data, event=event)

    def exchange_hid(self, cmd: int, addr: int = 0, data: bytes = b"A" * 16) -> bytes:
        req = NKFido2Client.format_request(cmd, addr, data)

        data = self.send_data_hid(SoloBootloader.HIDCommandBoot, req)

        ret = data[0]
        if ret != CtapError.ERR.SUCCESS:
            raise CtapError(ret)

        return data[1:]

    def exchange_u2f(self, cmd: int, addr: int = 0, data: bytes = b"A" * 16) -> bytes:
        appid = b"A" * 32
        chal = b"B" * 32

        req = NKFido2Client.format_request(cmd, addr, data)

        res = self.ctap1.authenticate(chal, appid, req)

        ret = res.signature[0]
        if ret != CtapError.ERR.SUCCESS:
            raise CtapError(ret)

        return res.signature[1:]

    def bootloader_version(self) -> Tuple[int, int, int]:
        data = self.exchange(SoloBootloader.version)
        if len(data) > 2:
            return (data[0], data[1], data[2])
        return (0, 0, data[0])

    def solo_version(self) -> Union[bytes, Tuple[int, int, int]]:
        try:
            return self.send_data_hid(0x61, b"")
        except CtapError:
            data = self.exchange(SoloExtension.version)
            return (data[0], data[1], data[2])

    def write_flash(self, addr: int, data: bytes) -> None:
        self.exchange(SoloBootloader.write, addr, data)

    def boot_pubkey(self) -> bytes:
        return self.exchange(SoloBootloader.boot_pubkey)

    def get_rng(self, num: int = 0) -> bytes:
        ret = self.send_data_hid(SoloBootloader.HIDCommandRNG, struct.pack("B", num))
        return ret

    def get_status(self, num: int = 0) -> bytes:
        ret = self.send_data_hid(SoloBootloader.HIDCommandStatus, struct.pack("B", num))
        # print(ret[:8])
        return ret

    def verify_flash(self, sig: bytes) -> None:
        """
        Tells device to check signature against application.  If it passes,
        the application will boot.
        Exception raises if signature fails.
        """
        self.exchange(SoloBootloader.done, 0, sig)

    def wink(self) -> None:
        self.send_data_hid(CTAPHID.WINK, b"")

    def reset(self) -> None:
        assert isinstance(self.ctap2, Ctap2)
        self.ctap2.reset()

    def enter_bootloader(self) -> None:
        """
        If Nitrokey is configured as Nitrokey hacker or something similar,
        this command will tell the token to boot directly to the bootloader
        so it can be reprogrammed
        """
        if self.exchange != self.exchange_hid:
            self.send_data_hid(CTAPHID.INIT, b"\x11\x11\x11\x11\x11\x11\x11\x11")
        self.send_data_hid(SoloBootloader.HIDCommandEnterBoot, b"")

    def enter_bootloader_or_die(self) -> None:
        try:
            self.enter_bootloader()
        # except OSError:
        #     pass
        except CtapError as e:
            if e.code == CtapError.ERR.INVALID_COMMAND:
                print(
                    "Could not switch into bootloader mode.  Please hold down the button for 2s while you plug token in."
                )
                sys.exit(1)
            else:
                raise (e)

    def is_bootloader(self) -> bool:
        try:
            self.bootloader_version()
            return True
        except CtapError as e:
            if e.code == CtapError.ERR.INVALID_COMMAND:
                pass
            else:
                raise (e)
        except Exception:
            # exception during bootloader version check, assume no bootloader
            # local_print("could not get bootloader version")
            pass
        return False

    def program_file(self, name: str) -> bytes:  # noqa: C901
        def parseField(f: str) -> bytes:
            return base64.b64decode(helpers.from_websafe(f).encode())

        firmware_file_data = None
        if name.lower().endswith(".json"):
            firmware_file_data = json.loads(open(name, "r").read())
            fw = parseField(firmware_file_data["firmware"])
            sig = None

            if "versions" in firmware_file_data:
                current = (0, 0, 0)
                try:
                    current = self.bootloader_version()  # noqa: F841
                except CtapError as e:
                    if e.code == CtapError.ERR.INVALID_COMMAND:
                        pass
                    else:
                        raise (e)
                sig = parseField(firmware_file_data["versions"][">2.5.3"]["signature"])

                if sig is None:
                    raise RuntimeError(
                        "Improperly formatted firmware file.  Could not match version."
                    )
            else:
                sig = parseField(firmware_file_data["signature"])

            ih = IntelHex()
            tmp = tempfile.NamedTemporaryFile(delete=False)
            tmp.write(fw)
            tmp.seek(0)
            tmp.close()
            ih.fromfile(tmp.name, format="hex")
        else:
            if not name.lower().endswith(".hex"):
                print('Warning, assuming "%s" is an Intel Hex file.' % name)
            sig = None
            ih = IntelHex()
            ih.fromfile(name, format="hex")

        if self.exchange == self.exchange_hid:
            chunk = 2048
        else:
            chunk = 240

        seg = ih.segments()[0]
        size = seg[1] - seg[0]
        total = 0
        t1 = time.time() * 1000
        print("erasing firmware...")
        for i in range(seg[0], seg[1], chunk):
            s = i
            ext = min(i + chunk, seg[1])
            data = ih.tobinarray(start=i, size=ext - s)
            self.write_flash(i, data)  # type: ignore[arg-type]
            total += chunk
            progress = total / float(size) * 100
            sys.stdout.write("updating firmware %.2f%%...\r" % progress)
        sys.stdout.write("updated firmware 100%             \r\n")
        t2 = time.time() * 1000
        print("time: %.2f s" % ((t2 - t1) / 1000.0))

        if sig is None:
            sig = b"A" * 64

        success = False
        if self.do_reboot:
            try:
                print("bootloader is verifying signature...")
                print(f"Trying with {sig.hex()}")
                self.verify_flash(sig)
                print("...pass!")
                success = True
            except CtapError as e:
                if e.code != 0x27:
                    raise
                print("...error!")

        if not success:
            msg = """Bootloader reports failure in the signature verification. If your device is staying in the
            bootloader mode after reinserting it into the USB port, please execute the following to make it work again:

            # Download the compressed firmware file for Nitrokey FIDO2 128 v2.4.1 and extract it
            wget https://github.com/Nitrokey/nitrokey-fido2-firmware/releases/download/2.4.1.nitrokey/nitrokey-fido2-firmware-2.4.1-128kB-app-signed.zip
            unzip nitrokey-fido2-firmware-2.4.1-128kB-app-signed.zip
            # Run the update process again with the just downloaded firmware
            nitropy fido2 util program bootloader nitrokey-fido2-firmware-2.4.1-128kB-app-signed.json
            """
            local_critical(msg, support_hint=False)

        return sig



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import base64
import contextlib
import datetime
import json
import logging
import mimetypes
import os
import os.path
import sys
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional, Protocol, Sequence

import click
import nethsm as nethsm_sdk
from click import Context
from nethsm import Authentication, Base64, ClusterJoinData, NetHSM, State
from nethsm.backup import EncryptedBackup

from pynitrokey.cli.exceptions import CliException
from pynitrokey.helpers import prompt


class EnumMeta(Protocol):
    def __iter__(self) -> Iterator[Enum]: ...


def make_enum_type(enum_cls: EnumMeta) -> click.Choice[Enum]:
    return click.Choice([variant.value for variant in enum_cls], case_sensitive=False)


def base64_input(s: str) -> Base64:
    return Base64.from_encoded(s, ignore_whitespace=True)


DATETIME_TYPE = click.DateTime(formats=["%Y-%m-%dT%H:%M:%S%z"])
ROLE_TYPE = make_enum_type(nethsm_sdk.Role)
LOG_LEVEL_TYPE = make_enum_type(nethsm_sdk.LogLevel)
UNATTENDED_BOOT_STATUS_TYPE = make_enum_type(nethsm_sdk.UnattendedBootStatus)
TYPE_TYPE = make_enum_type(nethsm_sdk.KeyType)
TYPE_TLS_KEY_TYPE = make_enum_type(nethsm_sdk.TlsKeyType)
MECHANISM_TYPE = make_enum_type(nethsm_sdk.KeyMechanism)
ENCRYPT_MODE_TYPE = make_enum_type(nethsm_sdk.EncryptMode)
DECRYPT_MODE_TYPE = make_enum_type(nethsm_sdk.DecryptMode)
SIGN_MODE_TYPE = make_enum_type(nethsm_sdk.SignMode)


def prompt_str(
    msg: str, default: Optional[str] = None, hide_input: bool = False
) -> str:
    value = prompt(msg, default=default, hide_input=hide_input)
    assert isinstance(value, str)
    return value


def print_row(values: Iterable[str], widths: Iterable[int]) -> None:
    row = [value.ljust(width) for (value, width) in zip(values, widths)]
    print(*row, sep="\t")


def print_table(headers: Sequence[str], data: Iterable[Sequence[Any]]) -> None:
    widths = [len(header) for header in headers]
    str_data = []
    for row in data:
        str_row = []
        for i in range(len(widths)):
            str_value = str(row[i])
            str_row.append(str_value)
            widths[i] = max(widths[i], len(str_value))
        str_data.append(str_row)

    print_row(headers, widths)
    print_row(["-" * width for width in widths], widths)
    for row in str_data:
        print_row(row, widths)


@dataclass
class Config:
    host: Optional[str]
    username: Optional[str]
    password: Optional[str]
    verify_tls: bool
    ca_certs: Optional[str]
    debug: bool


@click.group()
@click.option("-h", "--host", "host", help="Set the host of the NetHSM API")
@click.option("-u", "--username", "username", help="The NetHSM user name")
@click.option("-p", "--password", "password", help="The NetHSM password")
@click.option(
    "--verify-tls/--no-verify-tls",
    default=True,
    help="Whether to verify the TLS certificate of the NetHSM",
)
@click.option(
    "--ca-certs",
    help="Path to the CA certs to use for the TLS verification",
)
@click.option("--debug", is_flag=True, help="Enable debug log messages")
@click.pass_context
def nethsm(
    ctx: Context,
    host: Optional[str],
    username: Optional[str],
    password: Optional[str],
    verify_tls: bool,
    ca_certs: Optional[str],
    debug: bool,
) -> None:
    """Interact with NetHSM devices, see subcommands."""

    ctx.obj = Config(
        host=host,
        username=username,
        password=password,
        verify_tls=verify_tls,
        ca_certs=ca_certs,
        debug=debug,
    )


@contextlib.contextmanager
def connect(ctx: Context, require_auth: bool = True) -> Iterator[NetHSM]:
    config = ctx.obj
    assert isinstance(config, Config)

    host = config.host
    if host is None:
        v = "NETHSM_HOST"
        if v not in os.environ:
            raise CliException(
                "Missing NetHSM host: set the --host option or the "
                f"{v} environment variable",
                support_hint=False,
            )
        host = os.environ[v]

    auth = None
    if require_auth:
        username = config.username
        password = config.password
        if not username:
            username = prompt_str(f"[auth] User name for NetHSM {host}")
        if not password:
            password = prompt_str(
                f"[auth] Password for user {username} on NetHSM {host}",
                hide_input=True,
            )
        auth = Authentication(username=username, password=password)

    nethsm = NetHSM(
        host, auth=auth, verify_tls=config.verify_tls, ca_certs=config.ca_certs
    )
    if config.debug:
        loggers = ["nethsm.client", "urllib3"]
        for logger in loggers:
            logging.getLogger(logger).setLevel(logging.DEBUG)
    try:
        yield nethsm
    except nethsm_sdk.NetHSMError as e:
        raise click.ClickException(f"NetHSM request failed: {e}")
    except nethsm_sdk.NetHSMRequestError as e:
        if e.type == nethsm_sdk.RequestErrorType.SSL_ERROR:
            raise click.ClickException(
                f"Could not connect to the NetHSM: {e.reason}\nIf you use a self-signed certificate, please set the --no-verify-tls option."
            )
        else:
            raise click.ClickException(
                f"Cound not connect to the NetHSM: {e.reason}\nIs the NetHSM running and reachable?"
            )
    finally:
        nethsm.close()


@nethsm.command()
@click.argument("passphrase", required=False)
@click.pass_context
def unlock(ctx: Context, passphrase: Optional[str]) -> None:
    """Bring a locked NetHSM into operational state."""
    with connect(ctx, require_auth=False) as nethsm:
        if not passphrase:
            passphrase = prompt_str(
                f"Unlock passphrase for NetHSM {nethsm.host}", hide_input=True
            )
        nethsm.unlock(passphrase)
        print(f"NetHSM {nethsm.host} unlocked")


@nethsm.command()
@click.pass_context
def lock(ctx: Context) -> None:
    """Bring an operational NetHSM into locked state.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.lock()
        print(f"NetHSM {nethsm.host} locked")


@nethsm.command()
@click.option(
    "-u",
    "--unlock-passphrase",
    hide_input=True,
    confirmation_prompt=True,
    prompt=True,
    help="The unlock passphrase to set",
)
@click.option(
    "-a",
    "--admin-passphrase",
    hide_input=True,
    confirmation_prompt=True,
    prompt=True,
    help="The admin passphrase to set",
)
@click.option(
    "-t",
    "--system-time",
    type=DATETIME_TYPE,
    help="The system time to set (default: the time of this system)",
)
@click.pass_context
def provision(
    ctx: Context,
    unlock_passphrase: str,
    admin_passphrase: str,
    system_time: Optional[datetime.datetime],
) -> None:
    """Initial provisioning of a NetHSM.

    If the unlock or admin passphrases are not set, they have to be entered
    interactively.  If the system time is not set, the current system time is
    used."""
    if not system_time:
        system_time = datetime.datetime.now(datetime.timezone.utc)

    print(
        "Warning: The unlock passphrase cannot be reset without knowing the current value. If the "
        "unlock passphrase is lost, neither can it be reset to a new value nor can the NetHSM be "
        "unlocked.",
        file=sys.stderr,
    )

    with connect(ctx, require_auth=False) as nethsm:
        nethsm.provision(unlock_passphrase, admin_passphrase, system_time)
        print(f"NetHSM {nethsm.host} provisioned")


@nethsm.command()
@click.option(
    "--details/--no-details",
    default=True,
    help="Also query the real name and role of the user",
)
@click.pass_context
def list_users(ctx: Context, details: bool) -> None:
    """List all users on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        user_ids = nethsm.list_users()

        print(f"Users on NetHSM {nethsm.host}:")
        print()

        headers = ["User ID"]
        if details:
            headers += ["Real name", "Role"]
            data = []
            for user_id in user_ids:
                user = nethsm.get_user(user_id=user_id)
                data.append([user_id, user.real_name, user.role.value])
        else:
            data = [[user_id] for user_id in user_ids]

        print_table(headers, data)


@nethsm.command()
@click.argument("user-id")
@click.pass_context
def get_user(ctx: Context, user_id: str) -> None:
    """Query the real name and role for a user ID on the NetHSM.

    This command requires authentication as a user with the Administrator or
    Operator role."""
    with connect(ctx) as nethsm:
        user = nethsm.get_user(user_id=user_id)
        print(f"User {user_id} on NetHSM {nethsm.host}")
        print(f"Real name:  {user.real_name}")
        print(f"Role:       {user.role.value}")


@nethsm.command()
@click.option("-n", "--real-name", prompt=True, help="The real name of the new user")
@click.option(
    "-r", "--role", type=ROLE_TYPE, prompt=True, help="The role of the new user"
)
@click.option(
    "-p",
    "--passphrase",
    prompt=True,
    hide_input=True,
    confirmation_prompt=True,
    help="The passphrase of the new user",
)
@click.option("-u", "--user-id", help="The user ID of the new user")
@click.option("-N", "--namespace", help="The namespace of the new user")
@click.option(
    "--create-namespace",
    is_flag=True,
    help="Create the namespace after adding the user",
)
@click.pass_context
def add_user(
    ctx: Context,
    real_name: str,
    role: str,
    passphrase: str,
    user_id: Optional[str],
    namespace: Optional[str],
    create_namespace: bool,
) -> None:
    """Create a new user on the NetHSM.

    If the real name, role or passphrase are not specified, they have to be
    specified interactively.  If the user ID is not set, it is generated by the
    NetHSM.

    If a namespace is specified, the user will be created within the namespace.
    This means that the resulting user name will follow the pattern
    namespace~userid, i. e. the same user ID can be used in different
    namespaces.

    If the --create-namespace option is set and a namespace is specified, the
    namespace will be created after the user has been added.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        user_id = nethsm.add_user(
            real_name, nethsm_sdk.Role.from_string(role), passphrase, user_id, namespace
        )
        print(f"User {user_id} added to NetHSM {nethsm.host}")

        if namespace and nethsm.auth is not None and "~" not in nethsm.auth.username:
            # user added to non-existing namespace
            if create_namespace:
                nethsm.add_namespace(namespace)
                print(f"Namespace {namespace} added to NetHSM {nethsm.host}")
            else:
                print(
                    f"Warning: The namespace {namespace} does not exist.  Add it to the NetHSM with ",
                    file=sys.stderr,
                )
                print(f"    nitropy nethsm add-namespace {namespace}", file=sys.stderr)
                print(
                    "to be able to use it.  Once the namespace has been added, it can only be managed "
                    "by users in the same namespace.",
                    file=sys.stderr,
                )


@nethsm.command()
@click.argument("user-id")
@click.pass_context
def delete_user(ctx: Context, user_id: str) -> None:
    """Delete the user with the given user ID on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.delete_user(user_id)
        print(f"User {user_id} deleted on NetHSM {nethsm.host}")


@nethsm.command()
@click.pass_context
def list_namespaces(ctx: Context) -> None:
    """List all namespaces on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        namespaces = nethsm.list_namespaces()

        print(f"Namespaces on NetHSM {nethsm.host}:")
        for namespace in namespaces:
            print(f"- {namespace}")


@nethsm.command()
@click.argument("namespace")
@click.pass_context
def add_namespace(ctx: Context, namespace: str) -> None:
    """Add a new namespace on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.add_namespace(namespace)
        print(f"Namespace {namespace} added to NetHSM {nethsm.host}")


@nethsm.command()
@click.argument("namespace")
@click.pass_context
def delete_namespace(ctx: Context, namespace: str) -> None:
    """Delete a namespace on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.delete_namespace(namespace)
        print(f"Namespace {namespace} deleted on NetHSM {nethsm.host}")


@nethsm.command()
@click.option("-u", "--user-id", help="The user ID of the user")
@click.option(
    "-p",
    "--passphrase",
    prompt=True,
    hide_input=True,
    confirmation_prompt=True,
    help="The new passphrase of the user",
)
@click.pass_context
def set_passphrase(ctx: Context, user_id: Optional[str], passphrase: str) -> None:
    """Set the passphrase for the user with the given ID (or the current user).

    This command requires authentication as a user with the Administrator or
    Operator role.  Users with the Operator role can only change their own
    passphrase."""
    with connect(ctx) as nethsm:
        if not user_id:
            assert nethsm.auth is not None
            user_id = nethsm.auth.username
        nethsm.set_passphrase(user_id, passphrase)
        print(f"Updated the passphrase for user {user_id} on NetHSM {nethsm.host}")


@nethsm.command()
@click.argument("user-id")
@click.pass_context
def list_operator_tags(ctx: Context, user_id: str) -> None:
    """List the tags for an operator user ID on the NetHSM.

    This command requires authentication as a user with the Administrator role."""
    with connect(ctx) as nethsm:
        tags = nethsm.list_operator_tags(user_id=user_id)
        if tags:
            print(f"Tags for user {user_id}:")
            for tag in tags:
                print(f"- {tag}")
        else:
            print(f"No tags set for user {user_id}.")


@nethsm.command()
@click.argument("user-id")
@click.argument("tag")
@click.pass_context
def add_operator_tag(ctx: Context, user_id: str, tag: str) -> None:
    """Add a tag for an operator user on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.add_operator_tag(user_id=user_id, tag=tag)
        print(f"Added tag {tag} for user {user_id} on the NetHSM {nethsm.host}")


@nethsm.command()
@click.argument("user-id")
@click.argument("tag")
@click.pass_context
def delete_operator_tag(ctx: Context, user_id: str, tag: str) -> None:
    """Delete a tag for an operator user on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.delete_operator_tag(user_id=user_id, tag=tag)
        print(f"Deleted tag {tag} for user {user_id} on the NetHSM {nethsm.host}")


@nethsm.command()
@click.argument("key_id")
@click.argument("tag")
@click.pass_context
def add_key_tag(ctx: Context, key_id: str, tag: str) -> None:
    """Add a tag for a key on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.add_key_tag(key_id=key_id, tag=tag)
        print(f"Added tag {tag} for key {key_id} on the NetHSM {nethsm.host}")


@nethsm.command()
@click.argument("key_id")
@click.argument("tag")
@click.pass_context
def delete_key_tag(ctx: Context, key_id: str, tag: str) -> None:
    """Delete a tag for a key on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.delete_key_tag(key_id=key_id, tag=tag)
        print(f"Deleted tag {tag} for key {key_id} on the NetHSM {nethsm.host}")


@nethsm.command()
@click.pass_context
def info(ctx: Context) -> None:
    """Query the vendor and product information for a NetHSM."""
    with connect(ctx, require_auth=False) as nethsm:
        info = nethsm.get_info()
        print(f"Host:    {nethsm.host}")
        print(f"Vendor:  {info.vendor}")
        print(f"Product: {info.product}")


@nethsm.command()
@click.pass_context
def state(ctx: Context) -> None:
    """Query the state of a NetHSM."""
    with connect(ctx, require_auth=False) as nethsm:
        state = nethsm.get_state()
        print(f"NetHSM {nethsm.host} is {state.value}")


@nethsm.command()
@click.argument("length", type=int)
@click.pass_context
def random(ctx: Context, length: int) -> None:
    """Retrieve random bytes from the NetHSM as a Base64 string.

    This command requires authentication as a user with the Operator role."""
    with connect(ctx) as nethsm:
        print(nethsm.get_random_data(length))


@nethsm.command()
@click.pass_context
def metrics(ctx: Context) -> None:
    """Query the metrics of a NetHSM.

    This command requires authentication as a user with the Metrics role."""
    with connect(ctx) as nethsm:
        headers = ["Metric", "Value"]
        data = nethsm.get_metrics()
        print_table(headers, [list(row) for row in sorted(data.items())])


@nethsm.command()
@click.option(
    "--details/--no-details",
    default=True,
    help="Also query the key data",
)
@click.option(
    "-f",
    "--filter",
    type=str,
    help="Filter keys by tags for respective user",
)
@click.option("-p", "--prefix", type=str, help="Only list keys with the given prefix")
@click.pass_context
def list_keys(
    ctx: Context, details: bool, filter: Optional[str], prefix: Optional[str]
) -> None:
    """List all keys on the NetHSM.

    This command requires authentication as a user with the Administrator or
    Operator role."""
    with connect(ctx) as nethsm:
        key_ids = nethsm.list_keys(filter, prefix=prefix)

        print(f"Keys on NetHSM {nethsm.host}:")
        print()

        headers = ["Key ID"]
        if details:
            headers += ["Type", "Mechanisms", "Operations", "Tags"]
            data = []
            for key_id in key_ids:
                key = nethsm.get_key(key_id=key_id)
                data.append(
                    [
                        key_id,
                        key.type.value,
                        ", ".join([m.value for m in key.mechanisms]),
                        key.operations,
                        ", ".join(key.tags) if key.tags is not None else "",
                    ]
                )
        else:
            data = [[key_id] for key_id in key_ids]

        print_table(headers, data)


@nethsm.command()
@click.argument("key_id")
@click.option("--public-key", is_flag=True, help="Query the public key as a PEM file")
@click.pass_context
def get_key(ctx: Context, key_id: str, public_key: bool) -> None:
    """Get information about a key on the NetHSM.

    This command requires authentication as a user with the Administrator or
    Operator role."""
    with connect(ctx) as nethsm:
        if public_key:
            print(nethsm.get_key_public_key(key_id))
        else:
            key = nethsm.get_key(key_id)
            mechanisms = ", ".join([str(m.value) for m in key.mechanisms])
            print(f"Key {key_id} on NetHSM {nethsm.host}:")
            print(f"Type:            {key.type.value}")
            print(f"Mechanisms:      {mechanisms}")
            print(f"Operations:      {key.operations}")
            if key.tags:
                tags = ", ".join(key.tags)
                print(f"Tags:            {tags}")

            if isinstance(key.public_key, nethsm_sdk.RsaPublicKey):
                print(f"Modulus:         {key.public_key.modulus}")
                print(f"Public exponent: {key.public_key.public_exponent}")
            elif isinstance(key.public_key, nethsm_sdk.EcPublicKey):
                print(f"Data:            {key.public_key.data}")
            elif key.public_key is not None:
                print(f"Public key:      {key.public_key}")


@nethsm.command()
@click.argument("old-key-id")
@click.argument("new-key-id")
@click.pass_context
def move_key(ctx: Context, old_key_id: str, new_key_id: str) -> None:
    """Move the key pair with the given old key ID to the new key ID (requires NetHSM v3).

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.move_key(old_key_id, new_key_id)
        print(f"Key {old_key_id} moved to {new_key_id} on NetHSM {nethsm.host}")


@nethsm.command()
@click.argument("key-id")
@click.pass_context
def delete_key(ctx: Context, key_id: str) -> None:
    """Delete the key pair with the given key ID on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.delete_key(key_id)
        print(f"Key {key_id} deleted on NetHSM {nethsm.host}")


def prompt_mechanisms(type: str) -> list[str]:
    # We assume that key type X corresponds to the mechanisms starting with X.
    # This is no longer true for curves, so we have to adapt the type
    if type == nethsm_sdk.KeyType.CURVE25519.value:
        type = "EdDSA"
    elif type.startswith("EC_"):
        type = "ECDSA"

    available_mechanisms = []
    print("Supported mechanisms for this key type:")
    for mechanism in nethsm_sdk.KeyMechanism:
        if mechanism.value.startswith(type):
            available_mechanisms.append(mechanism.value)
            print(f"  {mechanism.value}")

    # If there is only one matching algorithm, we can choose it and don’t have
    # to ask the user.
    if len(available_mechanisms) == 1:
        print(f"Automatically selecting the key mechanism {available_mechanisms[0]}")
        return available_mechanisms

    print(
        "Please enter at least one mechanism.  Enter an empty string to "
        "finish the list of mechanisms."
    )

    mechanism_type = click.Choice(available_mechanisms, case_sensitive=False)
    mechanisms: list[str] = []
    cont = True
    while cont:
        default = None
        prompt_text = "Add mechanism"
        if mechanisms:
            prompt_text += " (or empty string to continue)"
            default = ""
        mechanism = prompt(
            prompt_text,
            type=mechanism_type,
            default=default,
            show_choices=False,
            show_default=False,
        )

        if "" not in available_mechanisms:
            available_mechanisms.append("")

        assert isinstance(mechanism, str)
        if mechanism:
            mechanisms.append(mechanism)
            available_mechanisms.remove(mechanism)
        else:
            cont = False

    if not mechanisms:
        raise click.ClickException("No key mechanisms selected!")

    return mechanisms


@nethsm.command()
@click.option(
    "-t",
    "--type",
    type=TYPE_TYPE,
    prompt=True,
    help="The type for the new key",
)
@click.option(
    "-m",
    "--mechanism",
    "mechanisms",
    type=MECHANISM_TYPE,
    multiple=True,
    help="The mechanisms for the new key",
)
@click.option(
    "--tags",
    type=str,
    multiple=True,
    help="The tags for the new key",
)
@click.option(
    "-p",
    "--prime-p",
    help="The prime p for RSA keys, base64-encoded",
)
@click.option(
    "-q",
    "--prime-q",
    help="The prime q for RSA keys, base64-encoded",
)
@click.option(
    "-e",
    "--public-exponent",
    help="The public exponent for RSA keys, base64-encoded",
)
@click.option(
    "-d",
    "--data",
    help="The key data for ED25519 or ECDSA_* keys, base64-encoded",
)
@click.option(
    "-k",
    "--key-id",
    help="The ID of the new key",
)
@click.pass_context
def add_key(
    ctx: Context,
    type: str,
    mechanisms: list[str],
    tags: list[str],
    prime_p: Optional[str],
    prime_q: Optional[str],
    public_exponent: Optional[str],
    data: Optional[str],
    key_id: Optional[str],
) -> None:
    """Add a key pair on the NetHSM.

    If the key ID is not set, it is generated by the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    key_type = nethsm_sdk.KeyType.from_string(type)
    mechanisms = list(mechanisms) or prompt_mechanisms(type)

    private_key: nethsm_sdk.PrivateKey
    if key_type == nethsm_sdk.KeyType.RSA:
        if data:
            raise click.ClickException("-d/--data must not be set for RSA keys")
        if not prime_p:
            prime_p = prompt_str("Prime p")
        if not prime_q:
            prime_q = prompt_str("Prime q")
        if not public_exponent:
            public_exponent = prompt_str("Public exponent")
        private_key = nethsm_sdk.RsaPrivateKey(
            prime_p=base64_input(prime_p),
            prime_q=base64_input(prime_q),
            public_exponent=base64_input(public_exponent),
        )
    else:
        if prime_p:
            raise click.ClickException("-p/--prime-p may only be set for RSA keys")
        if prime_q:
            raise click.ClickException("-q/--prime-q may only be set for RSA keys")
        if public_exponent:
            raise click.ClickException(
                "-e/--public-exponent may only be set for RSA keys"
            )
        if not data:
            data = prompt_str("Key data")
        private_key = nethsm_sdk.GenericPrivateKey(data=base64_input(data))

    with connect(ctx) as nethsm:
        key_id = nethsm.add_key(
            key_id=key_id,
            type=key_type,
            mechanisms=[nethsm_sdk.KeyMechanism.from_string(m) for m in mechanisms],
            tags=tags,
            private_key=private_key,
        )
        print(f"Key {key_id} added to NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-m",
    "--mechanism",
    "mechanisms",
    type=MECHANISM_TYPE,
    multiple=True,
    help="The mechanisms for the new key",
)
@click.option(
    "--tags",
    type=str,
    multiple=True,
    help="The tags for the new key",
)
@click.option(
    "-k",
    "--key-id",
    help="The ID of the new key",
)
@click.argument("filename")
@click.pass_context
def import_key(
    ctx: Context,
    mechanisms: list[str],
    tags: list[str],
    key_id: Optional[str],
    filename: str,
) -> None:
    """Import a key pair from a PEM file into the NetHSM.

    If the key ID is not set, it is generated by the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    mechanisms = list(mechanisms)

    with open(filename) as f:
        private_key = f.read()

    with connect(ctx) as nethsm:
        key_id = nethsm.add_key_pem(
            key_id=key_id,
            mechanisms=[nethsm_sdk.KeyMechanism.from_string(m) for m in mechanisms],
            tags=tags,
            private_key=private_key,
        )
        print(f"Key {key_id} added to NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "type",
    "-t",
    "--type",
    type=TYPE_TYPE,
    prompt=True,
    help="The type for the generated key",
)
@click.option(
    "-m",
    "--mechanism",
    "mechanisms",
    type=MECHANISM_TYPE,
    multiple=True,
    help="The mechanisms for the generated key",
)
@click.option(
    "-l",
    "--length",
    type=int,
    prompt=True,
    help="The length of the generated key",
)
@click.option(
    "-k",
    "--key-id",
    help="The ID of the generated key",
)
@click.pass_context
def generate_key(
    ctx: Context, type: str, mechanisms: list[str], length: int, key_id: Optional[str]
) -> None:
    """Generate a key pair on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    mechanisms = list(mechanisms) or prompt_mechanisms(type)
    with connect(ctx) as nethsm:
        key_id = nethsm.generate_key(
            nethsm_sdk.KeyType.from_string(type),
            [nethsm_sdk.KeyMechanism.from_string(m) for m in mechanisms],
            length,
            key_id,
        )
        print(f"Key {key_id} generated on NetHSM {nethsm.host}")


def _optional_or(s: Optional[str], default: str) -> str:
    if s is None:
        return default
    return s


@nethsm.command()
@click.option("--logging", is_flag=True, help="Query the logging configuration")
@click.option("--network", is_flag=True, help="Query the network configuration")
@click.option("--time", is_flag=True, help="Query the system time")
@click.option(
    "--unattended-boot", is_flag=True, help="Query the unattended boot configuration"
)
@click.option("--public-key", is_flag=True, help="Query the public key")
@click.option("--certificate", is_flag=True, help="Query the certificate")
@click.pass_context
def get_config(
    ctx: Context,
    logging: bool,
    network: bool,
    time: bool,
    unattended_boot: bool,
    public_key: bool,
    certificate: bool,
) -> None:
    """Query the configuration of a NetHSM.

    Only the configuration items selected with the corresponding option are
    queried.  If no option is set, all items are queried.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        print(f"Configuration for NetHSM {nethsm.host}:")
        show_all = not any(
            [logging, network, time, unattended_boot, public_key, certificate]
        )

        if show_all or logging:
            logging_config = nethsm.get_config_logging()
            print("  Logging:")
            print("    IP address:   ", logging_config.ip_address)
            print("    Port:         ", logging_config.port)
            print("    Log level:    ", logging_config.log_level)

        if show_all or network:
            network_config = nethsm.get_config_network()
            print("  Network:")
            print("    IP address:   ", network_config.ip_address)
            print("    Netmask:      ", network_config.netmask)
            print("    Gateway:      ", network_config.gateway)
            ipv6 = "not configured" if network_config.ipv6 is None else ""
            print("    IPv6:         ", ipv6)
            if network_config.ipv6 is not None:
                print("        CIDR:     ", network_config.ipv6.cidr)
                gateway = _optional_or(network_config.ipv6.gateway, "not configured")
                print("        Gateway:  ", gateway)

        if show_all or time:
            time_config = nethsm.get_config_time()
            print("  Time:           ", time_config)

        if show_all or unattended_boot:
            unattended_boot_config = nethsm.get_config_unattended_boot()
            print("  Unattended boot:", unattended_boot_config)

        if show_all or public_key:
            public_key_config = nethsm.get_public_key()
            print("  Public key:")
            for line in public_key_config.splitlines():
                print(f"    {line}")

        if show_all or certificate:
            certificate_config = nethsm.get_certificate()
            print("  Certificate:")
            for line in certificate_config.splitlines():
                print(f"    {line}")


@nethsm.command()
@click.option(
    "-n",
    "--new-passphrase",
    hide_input=True,
    confirmation_prompt=True,
    prompt=True,
    help="The new backup passphrase",
)
@click.option(
    "-p",
    "--current-passphrase",
    help="The current backup passphrase (or an empty string if not set)",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Do not ask for confirmation before changing the passphrase",
)
@click.pass_context
def set_backup_passphrase(
    ctx: Context, new_passphrase: str, current_passphrase: Optional[str], force: bool
) -> None:
    """Set the backup passphrase of a NetHSM.

    Changing the backup passphrase requires the current passphrase (if set,
    empty string otherwise).

    This command requires authentication as a user with the Administrator
    role."""

    print(
        "Warning: The backup passphrase cannot be reset without knowing the current value. If the "
        "backup passphrase is lost, neither can it be reset to a new value nor can the created "
        "backups be restored.",
        file=sys.stderr,
    )

    confirmed = force or click.confirm("Do you want to continue?")
    if not confirmed:
        raise click.Abort()

    if not current_passphrase:
        current_passphrase = prompt_str(
            "The current backup passphrase (or an empty string if not set)",
            hide_input=True,
            default="",
        )
    with connect(ctx) as nethsm:
        nethsm.set_backup_passphrase(
            new_passphrase=new_passphrase, current_passphrase=current_passphrase
        )
        print(f"Updated the backup passphrase for NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-n",
    "--new-passphrase",
    hide_input=True,
    confirmation_prompt=True,
    prompt=True,
    help="The new unlock passphrase",
)
@click.option(
    "-p",
    "--current-passphrase",
    hide_input=True,
    prompt=True,
    help="The current unlock passphrase",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Do not ask for confirmation before changing the passphrase",
)
@click.pass_context
def set_unlock_passphrase(
    ctx: Context, new_passphrase: str, current_passphrase: str, force: bool
) -> None:
    """Set the unlock passphrase of a NetHSM.

    Changing the unlock passphrase requires the current passphrase.

    This command requires authentication as a user with the Administrator
    role."""

    print(
        "Warning: The unlock passphrase cannot be reset without knowing the current value. If the "
        "unlock passphrase is lost, neither can it be reset to a new value nor can the NetHSM be "
        "unlocked.",
        file=sys.stderr,
    )

    confirmed = force or click.confirm("Do you want to continue?")
    if not confirmed:
        raise click.Abort()

    with connect(ctx) as nethsm:
        nethsm.set_unlock_passphrase(
            new_passphrase=new_passphrase, current_passphrase=current_passphrase
        )
        print(f"Updated the unlock passphrase for NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-a",
    "--ip-address",
    help="The IP address of the new logging destination",
    required=True,
)
@click.option(
    "-p",
    "--port",
    type=int,
    help="The port of the new logging destination",
    required=True,
)
@click.option(
    "-l",
    "--log-level",
    type=LOG_LEVEL_TYPE,
    help="The new log level",
    required=True,
)
@click.pass_context
def set_logging_config(
    ctx: Context, ip_address: str, port: int, log_level: str
) -> None:
    """Set the logging configuration of a NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.set_logging_config(
            ip_address, port, nethsm_sdk.LogLevel.from_string(log_level)
        )
        print(f"Updated the logging configuration for NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-a",
    "--ip-address",
    help="The new IP address",
    required=True,
)
@click.option(
    "-n",
    "--netmask",
    help="The new netmask",
    required=True,
)
@click.option(
    "-g",
    "--gateway",
    help="The new gateway",
    required=True,
)
@click.option("--ipv6-cidr")
@click.option("--ipv6-gateway")
@click.pass_context
def set_network_config(
    ctx: Context,
    ip_address: str,
    netmask: str,
    gateway: str,
    ipv6_cidr: Optional[str],
    ipv6_gateway: Optional[str],
) -> None:
    """Set the network configuration of a NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    ipv6 = None
    if ipv6_cidr is not None or ipv6_gateway is not None:
        if ipv6_cidr is None:
            raise CliException(
                f"--ipv6-cidr must be set if --ipv6-gateway is set", support_hint=False
            )
        ipv6 = nethsm_sdk.Ipv6Config(cidr=ipv6_cidr, gateway=ipv6_gateway)
    with connect(ctx) as nethsm:
        nethsm.set_network_config(ip_address, netmask, gateway, ipv6=ipv6)
        print(f"Updated the network configuration for NetHSM {nethsm.host}")


@nethsm.command()
@click.argument(
    "time",
    type=DATETIME_TYPE,
    required=False,
)
@click.pass_context
def set_time(ctx: Context, time: Optional[datetime.datetime]) -> None:
    """Set the system time of a NetHSM.

    If the time is not given as an argument, the system time of this system is used.

    This command requires authentication as a user with the Administrator
    role."""
    if not time:
        time = datetime.datetime.now(datetime.timezone.utc)
    with connect(ctx) as nethsm:
        nethsm.set_time(time)
        print(f"Updated the system time for NetHSM {nethsm.host}")


@nethsm.command()
@click.argument(
    "status",
    type=UNATTENDED_BOOT_STATUS_TYPE,
)
@click.pass_context
def set_unattended_boot(ctx: Context, status: str) -> None:
    """Set the unattended boot configuration of a NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.set_unattended_boot(nethsm_sdk.UnattendedBootStatus.from_string(status))
        print(f"Updated the unattended boot configuration for NetHSM {nethsm.host}")


def get_api_or_key_id(api: bool, key_id: Optional[str]) -> tuple[bool, Optional[str]]:
    """Helper method for operations that can be executed either for the API
    certificate or for the certificate stored for a key."""
    if api and key_id:
        raise click.ClickException("--api and --key-id are mutually exclusive")

    if not api and not key_id:
        choice = click.Choice(["api", "key"], case_sensitive=False)
        method = prompt(
            "For stored key or for NetHSM TLS interface?",
            type=choice,
        )
        if method == "api":
            api = True
        elif method == "key":
            key_id = prompt("Key ID")
        else:
            raise ValueError("Unexpected method")

    return (api, key_id)


@nethsm.command()
@click.option(
    "-a", "--api", is_flag=True, help="Set the certificate for the NetHSM TLS interface"
)
@click.option("-k", "--key-id", help="The ID of the key to set the certificate for")
@click.argument("filename")
@click.pass_context
def set_certificate(
    ctx: Context, api: bool, key_id: Optional[str], filename: str
) -> None:
    """Set a certificate on the NetHSM.

    If the --api option is set, the certificate used for the NetHSM TLS interface
    is set.  If the --key-id option is set, the certificate for a key stored on
    the NetHSM is set.

    This command requires authentication as a user with the Administrator
    role."""
    api, key_id = get_api_or_key_id(api, key_id)
    with connect(ctx) as nethsm:
        with open(filename, "rb") as f:
            if key_id:
                nethsm.set_key_certificate(key_id, f)
                print(
                    f"Updated the certificate for key {key_id} on NetHSM {nethsm.host}"
                )
            else:
                nethsm.set_certificate(f)
                print(f"Updated the API certificate for NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-a", "--api", is_flag=True, help="Get the certificate for the NetHSM TLS interface"
)
@click.option("-k", "--key-id", help="The ID of the key to get the certificate for")
@click.pass_context
def get_certificate(ctx: Context, api: bool, key_id: Optional[str]) -> None:
    """Get a certificate from the NetHSM.

    If the --api option is set, the certificate used for the NetHSM TLS interface
    is queried.  If the --key-id option is set, the certificate for a key stored on
    the NetHSM is queried.

    This command requires authentication as a user with the Administrator role.
    The certificate for a key can also be queried by a user with the Operator
    role."""
    api, key_id = get_api_or_key_id(api, key_id)
    with connect(ctx) as nethsm:
        if key_id:
            cert = nethsm.get_key_certificate(key_id)
            try:
                print(cert.decode())
            except UnicodeError:
                print(cert)
        else:
            print(nethsm.get_certificate())


@nethsm.command()
@click.option(
    "-k",
    "--key-id",
    prompt=True,
    help="The ID of the key to delete the certificate for",
)
@click.pass_context
def delete_certificate(ctx: Context, key_id: str) -> None:
    """Delete a certificate for a stored key from the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.delete_key_certificate(key_id)
        print(f"Deleted certificate for key {key_id} on NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-a", "--api", is_flag=True, help="Generate a CSR for the NetHSM TLS interface"
)
@click.option("-k", "--key-id", help="The ID of the key to generate the CSR for")
@click.option("--country", default="", prompt=True, help="The country name")
@click.option(
    "--state-or-province", default="", prompt=True, help="The state or province name"
)
@click.option("--locality", default="", prompt=True, help="The locality name")
@click.option("--organization", default="", prompt=True, help="The organization name")
@click.option(
    "--organizational-unit", default="", prompt=True, help="The organization unit name"
)
@click.option("--common-name", default="", prompt=True, help="The common name")
@click.option("--email-address", default="", prompt=True, help="The email address")
@click.pass_context
def csr(
    ctx: Context,
    api: bool,
    key_id: Optional[str],
    country: str,
    state_or_province: str,
    locality: str,
    organization: str,
    organizational_unit: str,
    common_name: str,
    email_address: str,
) -> None:
    """Generate a certificate signing request.

    If the --api option is set, the CSR is generated for the NetHSM, for
    example to replace the self-signed initial certificate.  If the --key-id
    option is set, the CSR is generated for a key stored on the NetHSM.

    This command requires authentication as a user with the Administrator
    role."""
    api, key_id = get_api_or_key_id(api, key_id)
    with connect(ctx) as nethsm:
        if key_id:
            csr = nethsm.key_csr(
                key_id=key_id,
                country=country,
                state_or_province=state_or_province,
                locality=locality,
                organization=organization,
                organizational_unit=organizational_unit,
                common_name=common_name,
                email_address=email_address,
            )
        else:
            csr = nethsm.csr(
                country=country,
                state_or_province=state_or_province,
                locality=locality,
                organization=organization,
                organizational_unit=organizational_unit,
                common_name=common_name,
                email_address=email_address,
            )
        print(csr)


@nethsm.command()
@click.option(
    "type",
    "-t",
    "--type",
    type=TYPE_TLS_KEY_TYPE,
    prompt=True,
    help="The type for the generated key",
)
@click.option(
    "-l",
    "--length",
    type=int,
    help="The length of the generated key",
)
@click.pass_context
def generate_tls_key(ctx: Context, type: str, length: Optional[int]) -> None:
    """Generate key pair for NetHSM TLS interface.

    This command requires authentication as a user with the Administrator
    role."""
    key_type = nethsm_sdk.TlsKeyType.from_string(type)
    if key_type == nethsm_sdk.TlsKeyType.RSA:
        if not length:
            length = click.prompt("Length", type=int)
    else:
        if length:
            raise click.ClickException("-l/--length may only be set for RSA keys")

    with connect(ctx) as nethsm:
        nethsm.generate_tls_key(key_type, length)
        print(f"Key for TLS interface generated on NetHSM {nethsm.host}")


@nethsm.command()
@click.pass_context
def system_info(ctx: Context) -> None:
    """Get system information for a NetHSM instance.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        info = nethsm.get_system_info()
        print(f"Host:             {nethsm.host}")
        print(f"Firmware version: {info.firmware_version}")
        print(f"Software version: {info.software_version}")
        print(f"Hardware version: {info.hardware_version}")
        print(f"Build tag:        {info.build_tag}")
        if info.tpm.attestation_keys:
            print("Attestation keys")
            for key, value in info.tpm.attestation_keys.items():
                print(f"  {key}:           {value}")
        if info.tpm.platform_configuration_registers:
            print("Platform Configuration Registers")
            for key, value in info.tpm.platform_configuration_registers.items():
                print(f"  {key}:              {value}")


@nethsm.command()
@click.argument("filename")
@click.pass_context
def backup(ctx: Context, filename: str) -> None:
    """Make a backup of a NetHSM instance and write it to a file.

    This command requires authentication as a user with the Backup role."""
    if os.path.exists(filename):
        raise click.ClickException(f"Backup file {filename} already exists")
    with connect(ctx) as nethsm:
        data = nethsm.backup()
        with open(filename, "xb") as f:
            f.write(data)
            print(f"Backup for {nethsm.host} written to {filename}")
        try:
            EncryptedBackup.parse(data)
        except ValueError as e:
            raise CliException(f"Failed to validate backup: {e}", support_hint=False)


@nethsm.command()
@click.option(
    "-p",
    "--backup-passphrase",
    hide_input=True,
    prompt=True,
    help="The backup passphrase",
)
@click.option(
    "-t",
    "--system-time",
    type=DATETIME_TYPE,
    help="The system time to set (default: the time of this system)",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Restore the backup even if validation fails",
)
@click.argument("filename")
@click.pass_context
def restore(
    ctx: Context,
    backup_passphrase: str,
    system_time: Optional[datetime.datetime],
    force: bool,
    filename: str,
) -> None:
    """Restore a backup of a NetHSM instance from a file.

    If the system time is not set, the current system time is used."""
    if not system_time:
        system_time = datetime.datetime.now(datetime.timezone.utc)

    with open(filename, "rb") as f:
        data = f.read()
    try:
        EncryptedBackup.parse(data).decrypt(backup_passphrase)
    except ValueError as e:
        if force:
            print(f"Failed to validate backup: {e}")
            print("Backup is restored anyway as --force is set")
        else:
            raise CliException(
                f"Failed to validate backup (use --force to restore anyway): {e}",
                support_hint=False,
            )

    require_auth = False
    with connect(ctx, require_auth=False) as nethsm:
        state = nethsm.get_state()
        if state == State.OPERATIONAL:
            require_auth = True

    with connect(ctx, require_auth=require_auth) as nethsm:
        with open(filename, "rb") as f:
            nethsm.restore(data, backup_passphrase, system_time)
        print(f"Backup restored on NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-p",
    "--backup-passphrase",
    help="The backup passphrase for decryption (default: only the unencrypted metadata is validated)",
)
@click.argument("filename")
def validate_backup(backup_passphrase: Optional[str], filename: str) -> None:
    """Validate a NetHSM backup file.

    Per default, only the metadata of the encrypted backup is validated.  If
    the backup passphrase is set, the backup is decrypted and the content is
    also validated."""

    with open(filename, "rb") as f:
        data = f.read()
    try:
        encrypted = EncryptedBackup.parse(data)
    except ValueError as e:
        raise CliException(
            f"Failed to validate backup metadata: {e}", support_hint=False
        )

    if backup_passphrase:
        try:
            encrypted.decrypt(backup_passphrase)
        except ValueError as e:
            raise CliException(
                f"Failed to validate backup content: {e}", support_hint=False
            )
        print("Backup metadata and content are valid.")
    else:
        print("Backup metadata is valid.")


@nethsm.command()
@click.option(
    "-p",
    "--backup-passphrase",
    hide_input=True,
    prompt=True,
    help="The backup passphrase",
)
@click.argument("filename")
def export_backup(backup_passphrase: str, filename: str) -> None:
    """Export the content of a NetHSM backup file.

    The key-value data stored in the backup file is printed to the standard
    output as a JSON object using the base64 encoding for binary data.
    Additionally, the .locked-domain-key and .version keys are set with the
    domain key and version info extracted from the backup file."""

    with open(filename, "rb") as f:
        data = f.read()
    try:
        encrypted = EncryptedBackup.parse(data)
    except ValueError as e:
        raise CliException(f"Failed to parse backup metadata: {e}", support_hint=False)

    try:
        decrypted = encrypted.decrypt(backup_passphrase)
    except ValueError as e:
        raise CliException(f"Failed to decrypt backup content: {e}", support_hint=False)

    values: dict[str, Any] = {}
    values[".locked-domain-key"] = base64.b64encode(decrypted.domain_key).decode()
    values[".version"] = decrypted.version
    for key, value in decrypted.data.items():
        values[key] = base64.b64encode(value).decode()

    json.dump(values, sys.stdout, indent=4)
    print()


@nethsm.command()
@click.argument("filename")
@click.pass_context
def update(ctx: Context, filename: str) -> None:
    """Load an update to a NetHSM instance.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        with open(filename, "rb") as f:
            release_notes = nethsm.update(f)
        print(f"Image {filename} uploaded to NetHSM {nethsm.host}")
        if release_notes:
            print("Release notes:")
            print("  " + release_notes)


@nethsm.command()
@click.pass_context
def cancel_update(ctx: Context) -> None:
    """Cancel a queued update on a NetHSM instance.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.cancel_update()
        print(f"Update successfully cancelled on NetHSM {nethsm.host}")


@nethsm.command()
@click.pass_context
def commit_update(ctx: Context) -> None:
    """Commit a queued update on a NetHSM instance.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        nethsm.commit_update()
        print(f"Update successfully committed on NetHSM {nethsm.host}")


@nethsm.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force reboot",
)
@click.pass_context
def reboot(ctx: Context, force: bool) -> None:
    """Reboot a NetHSM instance.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        print(f"NetHSM {nethsm.host} will be rebooted.")
        reboot = force or click.confirm("Do you want to continue?")

        if reboot:
            nethsm.reboot()
            print(f"NetHSM {nethsm.host} is about to reboot")
        else:
            print(f"Reboot on NetHSM {nethsm.host} cancelled")


@nethsm.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force shutdown",
)
@click.pass_context
def shutdown(ctx: Context, force: bool) -> None:
    """Shutdown a NetHSM instance.

    This command requires authentication as a user with the Administrator
    role."""
    require_auth = False
    with connect(ctx, require_auth=require_auth) as nethsm:
        state = nethsm.get_state()
        if state == State.OPERATIONAL:
            require_auth = True

    with connect(ctx, require_auth=require_auth) as nethsm:
        print(f"NetHSM {nethsm.host} will be shutdown.")
        shutdown = force or click.confirm("Do you want to continue?")

        if shutdown:
            nethsm.shutdown()
            print(f"NetHSM {nethsm.host} is about to shutdown")
        else:
            print(f"Shutdown on NetHSM {nethsm.host} cancelled")


@nethsm.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force factory reset",
)
@click.pass_context
def factory_reset(ctx: Context, force: bool) -> None:
    """Perform a factory reset for a NetHSM instance.

    This command requires authentication as a user with the Administrator
    role."""
    with connect(ctx) as nethsm:
        print(f"NetHSM {nethsm.host} will be set to factory defaults.")
        print(f"All data will be lost!")
        factory_reset = force or click.confirm("Do you want to continue?")

        if factory_reset:
            nethsm.factory_reset()
            print(f"NetHSM {nethsm.host} is about to perform a factory reset")
        else:
            print(f"Factory reset on NetHSM {nethsm.host} cancelled")


@nethsm.command()
@click.option(
    "-k",
    "--key-id",
    prompt=True,
    help="The ID of the key to encrypt the data with",
)
@click.option(
    "-d",
    "--data",
    prompt=True,
    help="The data in Base64 encoding",
)
@click.option(
    "-m",
    "--mode",
    type=ENCRYPT_MODE_TYPE,
    prompt=True,
    help="The encrypt mode",
)
@click.option(
    "-iv",
    "--initialization-vector",
    "iv",
    type=str,
    help="The initialization vector in Base64 encoding",
)
@click.pass_context
def encrypt(ctx: Context, key_id: str, data: str, mode: str, iv: Optional[str]) -> None:
    """Encrypt data with an asymmetric secret key on the NetHSM and print the encrypted message.

    This command requires authentication as a user with the Operator role."""
    with connect(ctx) as nethsm:
        encrypted = nethsm.encrypt(
            key_id,
            base64_input(data),
            nethsm_sdk.EncryptMode.from_string(mode),
            iv=base64_input(iv) if iv else None,
        )
        print(f"Encrypted: {encrypted.encrypted.data}")
        print(f"Initialization vector: {encrypted.iv.data}")


@nethsm.command()
@click.option(
    "-k",
    "--key-id",
    prompt=True,
    help="The ID of the key to decrypt the data width",
)
@click.option(
    "-d",
    "--data",
    prompt=True,
    help="The encrypted data in Base64 encoding",
)
@click.option(
    "-m",
    "--mode",
    type=DECRYPT_MODE_TYPE,
    prompt=True,
    help="The decrypt mode",
)
@click.option(
    "-iv",
    "--initialization-vector",
    "iv",
    type=str,
    help="The initialization vector in Base64 encoding",
)
@click.pass_context
def decrypt(ctx: Context, key_id: str, data: str, mode: str, iv: Optional[str]) -> None:
    """Decrypt data with a secret key on the NetHSM and print the decrypted message.

    This command requires authentication as a user with the Operator role."""
    with connect(ctx) as nethsm:
        decrypted = nethsm.decrypt(
            key_id,
            base64_input(data),
            nethsm_sdk.DecryptMode.from_string(mode),
            base64_input(iv) if iv else None,
        )
        print(decrypted.data)


@nethsm.command()
@click.option(
    "-k",
    "--key-id",
    prompt=True,
    help="The ID of the key to sign the data width",
)
@click.option(
    "-d",
    "--data",
    prompt=True,
    help="The data to sign encoded using Base64",
)
@click.option(
    "-m",
    "--mode",
    type=SIGN_MODE_TYPE,
    prompt=True,
    help="The sign mode",
)
@click.pass_context
def sign(ctx: Context, key_id: str, data: str, mode: str) -> None:
    """Sign data with a secret key on the NetHSM and print the signature.

    This command requires authentication as a user with the Operator role."""
    with connect(ctx) as nethsm:
        signature = nethsm.sign(
            key_id, base64_input(data), nethsm_sdk.SignMode.from_string(mode)
        )
        print(signature.data)


@nethsm.command()
@click.option(
    "-a",
    "--admin-passphrase",
    default="adminadmin",
    help="The admin passphrase to set.",
)
@click.option(
    "-o",
    "--operator-passphrase",
    default="operatoroperator",
    help="The operator passphrase to set.",
)
@click.option(
    "-u",
    "--unlock-passphrase",
    default="unlockunlock",
    help="The unlock passphrase to set.",
)
@click.option(
    "-k",
    "--key-generations",
    default=1000,
    help="Number of keys to generate.",
)
@click.option(
    "-r",
    "--random-length",
    default=1024,
    help="Number of random bytes to generate.",
)
@click.option(
    "--skip-factory-reset",
    is_flag=True,
    help="Skip the factory reset at the end of the test.",
)
@click.pass_context
def test(
    ctx: Context,
    admin_passphrase: str,
    operator_passphrase: str,
    unlock_passphrase: str,
    key_generations: int,
    random_length: int,
    skip_factory_reset: bool,
) -> None:
    """Test a NetHSM by running certain operations."""

    admin_username = "admin"
    operator_username = "operator"

    with connect(ctx, require_auth=False) as nethsm:
        state = nethsm.get_state()
        if state == State.UNPROVISIONED:
            print("Unprovisioned NetHSM found.")
        else:
            raise CliException(
                "Provisioned or operational NetHSM found.", support_hint=False
            )

        system_time = datetime.datetime.now(datetime.timezone.utc)
        nethsm.provision(unlock_passphrase, admin_passphrase, system_time)
        print("Provisioned the NetHSM.")
        print(f"  Admin username: {admin_username}")
        print(f"  Admin passphrase: {admin_passphrase}")
        print(f"  Unlock passphrase: {unlock_passphrase}")

    ctx.obj.username = admin_username
    ctx.obj.password = admin_passphrase

    with connect(ctx, require_auth=True) as nethsm:
        nethsm.add_user(
            operator_username,
            nethsm_sdk.Role.OPERATOR,
            operator_passphrase,
            operator_username,
            None,
        )
        print("Created operator user.")
        print(f"  Operator username: {operator_username}")
        print(f"  Operator passphrase: {operator_passphrase}")

        for i in range(key_generations):
            nethsm.generate_key(
                nethsm_sdk.KeyType.RSA,
                [nethsm_sdk.KeyMechanism.RSA_DECRYPTION_RAW],
                2048,
                f"key{i}",
            )
        keys = len(nethsm.list_keys())
        if keys == key_generations:
            print(f"Generated {key_generations} RSA2048 keys.")
        else:
            raise CliException(
                f"The amount of keys requested and stored does not match.",
                support_hint=False,
            )

    ctx.obj.username = operator_username
    ctx.obj.password = operator_passphrase

    with connect(ctx, require_auth=True) as nethsm:
        MAX_RANDOM_LENGTH_PER_REQUEST = 1024
        received_random_length = 0
        remaining_random_length = random_length
        while remaining_random_length > 0:
            current_length = min(remaining_random_length, MAX_RANDOM_LENGTH_PER_REQUEST)
            data = nethsm.get_random_data(current_length)
            received_random_length += len(data.decode())
            remaining_random_length -= current_length
        if received_random_length == random_length:
            print(f"Generated {random_length} bytes of random data.")
        else:
            raise CliException(
                "The amount of random data requested and received does not match.",
                support_hint=False,
            )

    if not skip_factory_reset:
        ctx.obj.username = admin_username
        ctx.obj.password = admin_passphrase

        with connect(ctx, require_auth=True) as nethsm:
            print("Perform factory reset.")
            nethsm.factory_reset()


@nethsm.command()
@click.option(
    "--url",
    multiple=True,
)
@click.argument("join_data_path", type=Path)
@click.pass_context
def add_cluster_member(ctx: Context, url: tuple[str], join_data_path: Path) -> None:
    with connect(ctx) as nethsm:
        join_data = nethsm.add_cluster_member(list(url))

    s = json.dumps(join_data.to_dict())
    join_data_path.write_text(s)
    print("Wrote join data to {join_data_path}")


@nethsm.command()
@click.pass_context
def list_cluster_members(ctx: Context) -> None:
    with connect(ctx) as nethsm:
        cluster_members = nethsm.list_cluster_members()

    n = len(cluster_members)
    print(f"{n} cluster members:")
    for m in cluster_members:
        print(f"- id: {m.id}")
        print(f"  name: {m.name}")
        print(f"  peer URLs: {m.urls}")


@nethsm.command()
@click.option(
    "--url",
    multiple=True,
)
@click.argument("member-id")
@click.pass_context
def set_cluster_member_urls(ctx: Context, member_id: str, url: tuple[str]) -> None:
    with connect(ctx) as nethsm:
        nethsm.set_cluster_member_urls(member_id, list(url))


@nethsm.command()
@click.argument("member-id")
@click.pass_context
def remove_cluster_member(ctx: Context, member_id: str) -> None:
    with connect(ctx) as nethsm:
        nethsm.remove_cluster_member(member_id)


@nethsm.command()
@click.option("--backup-passphrase")
@click.argument("join_data_path", type=Path)
@click.pass_context
def join_cluster(ctx: Context, backup_passphrase: str, join_data_path: Path) -> None:
    data = json.loads(join_data_path.read_text())
    if not isinstance(data, dict):
        raise ValueError("Join data must contain an object")
    join_data = ClusterJoinData.from_dict(data)
    with connect(ctx) as nethsm:
        nethsm.join_cluster(data=join_data, backup_passphrase=backup_passphrase)


@nethsm.command()
@click.pass_context
def get_cluster_ca_certificate(ctx: Context) -> None:
    with connect(ctx) as nethsm:
        print(nethsm.get_cluster_ca_certificate())


@nethsm.command()
@click.argument("filename")
@click.pass_context
def set_cluster_ca_certificate(ctx: Context, filename: str) -> None:
    with connect(ctx) as nethsm:
        with open(filename, "rb") as f:
            nethsm.set_cluster_ca_certificate(f)



# Copyright 2019 SoloKeys Developers
# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT


class BasePyNKException(Exception):
    pass


class NonUniqueDeviceError(Exception):
    """When specifying a potentially destructive command...

    we check that either there is exactly one applicable device,
    or demand passing the serial number (same for ST DFU bootloader
    and Nitrokey bootloader+firmware.
    """

    pass


class NoSoloFoundError(Exception):
    """Can signify no Solo, or missing udev rule on Linux."""

    pass



# Copyright Nitrokey GmbH
# SPDX-License-Identifier: Apache-2.0 OR MIT

import datetime
import sys
from typing import Any, Iterable, Optional, Sequence, Tuple, Union

import click
import cryptography
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives._asymmetric import AsymmetricPadding
from cryptography.hazmat.primitives.asymmetric import ec, rsa
from cryptography.hazmat.primitives.asymmetric import utils as asym_utils
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.serialization import Encoding, pkcs12

from pynitrokey.cli.nk3 import nk3
from pynitrokey.helpers import check_experimental_flag, local_critical, local_print
from pynitrokey.tlv import Tlv

# Pyscard does not have wheels for all targets, leading to installation errors
# It is therefore made optional
#
# C901: `TryExcept` is too complex
try:  # noqa: C901
    from pynitrokey.nk3.piv_app import PivApp, find_by_id

    default_admin_key = "010203040506070801020304050607080102030405060708"
    all_key_ids = [
        "9A",
        "9C",
        "9D",
        "9E",
        "82",
        "83",
        "84",
        "85",
        "86",
        "87",
        "88",
        "89",
        "8A",
        "8B",
        "8C",
        "8D",
        "8E",
        "8F",
        "90",
        "91",
        "92",
        "93",
        "94",
        "95",
    ]

    key_id_click_type = type = click.Choice(
        [
            "9A",
            "9C",
            "9D",
            "9E",
            "82",
            "83",
            "84",
            "85",
            "86",
            "87",
            "88",
            "89",
            "8A",
            "8B",
            "8C",
            "8D",
            "8E",
            "8F",
            "90",
            "91",
            "92",
            "93",
            "94",
            "95",
        ],
        case_sensitive=False,
    )

    default_key_id = "9A"

    class RsaPivSigner(rsa.RSAPrivateKey):
        _device: PivApp
        _key_reference: int
        _public_key: rsa.RSAPublicKey

        def __init__(
            self, device: PivApp, key_reference: int, public_key: rsa.RSAPublicKey
        ):
            self._device = device
            self._key_reference = key_reference
            self._public_key = public_key

        def public_key(self) -> rsa.RSAPublicKey:
            return self._public_key

        @property
        def key_size(self) -> int:
            return self._public_key.key_size

        def sign(
            self,
            data: bytes,
            padding: AsymmetricPadding,
            # cryptography v47 adds NoDigestInfo to the union
            algorithm: Union[asym_utils.Prehashed, hashes.HashAlgorithm],  # type: ignore[override, unused-ignore]
        ) -> bytes:
            assert not isinstance(algorithm, asym_utils.Prehashed)
            assert isinstance(padding, PKCS1v15)
            assert isinstance(algorithm, hashes.SHA256)

            bits = self.key_size  # 2048 or 3072 or 4096
            return self._device.sign_rsa(data, self._key_reference, bits)

        def decrypt(self, ciphertext: bytes, padding: AsymmetricPadding) -> bytes:
            raise NotImplementedError()

        def private_numbers(self) -> rsa.RSAPrivateNumbers:
            raise NotImplementedError()

        def private_bytes(
            self,
            encoding: serialization.Encoding,
            format: serialization.PrivateFormat,
            encryption_algorithm: serialization.KeySerializationEncryption,
        ) -> bytes:
            raise NotImplementedError()

        def __copy__(self) -> "RsaPivSigner":
            raise NotImplementedError()

        def __deepcopy__(self, memo: dict[Any, Any]) -> "RsaPivSigner":
            raise NotImplementedError()

    class P256PivSigner(ec.EllipticCurvePrivateKey):
        _device: PivApp
        _key_reference: int
        _public_key: ec.EllipticCurvePublicKey

        def __init__(
            self,
            device: PivApp,
            key_reference: int,
            public_key: ec.EllipticCurvePublicKey,
        ):
            self._device = device
            self._key_reference = key_reference
            self._public_key = public_key

        def exchange(
            self, algorithm: ec.ECDH, peer_public_key: ec.EllipticCurvePublicKey
        ) -> bytes:
            raise NotImplementedError()

        def public_key(self) -> ec.EllipticCurvePublicKey:
            return self._public_key

        @property
        def curve(self) -> ec.EllipticCurve:
            return self._public_key.curve

        def private_numbers(self) -> ec.EllipticCurvePrivateNumbers:
            raise NotImplementedError()

        @property
        def key_size(self) -> int:
            return self._public_key.key_size

        def private_bytes(
            self,
            encoding: serialization.Encoding,
            format: serialization.PrivateFormat,
            encryption_algorithm: serialization.KeySerializationEncryption,
        ) -> bytes:
            raise NotImplementedError()

        def sign(
            self, data: bytes, signature_algorithm: ec.EllipticCurveSignatureAlgorithm
        ) -> bytes:
            assert isinstance(signature_algorithm, ec.ECDSA)
            assert isinstance(signature_algorithm.algorithm, hashes.SHA256)

            return self._device.sign_p256(data, self._key_reference)

        def __copy__(self) -> "P256PivSigner":
            raise NotImplementedError()

        def __deepcopy__(self, memo: dict[Any, Any]) -> "P256PivSigner":
            raise NotImplementedError()

    def print_row(values: Iterable[str], widths: Iterable[int]) -> None:
        row = [value.ljust(width) for (value, width) in zip(values, widths)]
        print(*row, sep="\t")

    def print_table(headers: Sequence[str], data: Iterable[Sequence[Any]]) -> None:
        widths = [len(header) for header in headers]
        str_data = []
        for row in data:
            str_row = []
            for i in range(len(widths)):
                str_value = str(row[i])
                str_row.append(str_value)
                widths[i] = max(widths[i], len(str_value))
            str_data.append(str_row)

        print_row(headers, widths)
        print_row(["-" * width for width in widths], widths)
        for row in str_data:
            print_row(row, widths)

    def import_rsa2048(
        device: PivApp,
        key_ref: int,
        key: rsa.RSAPrivateNumbers,
        public_key: rsa.RSAPublicNumbers,
    ) -> None:

        device.send_receive(
            0xFE,
            0x07,
            key_ref,
            Tlv.build(
                [
                    (0x01, key.p.to_bytes(256, "big")),
                    (0x02, key.q.to_bytes(256, "big")),
                    (
                        0x03,
                        public_key.e.to_bytes(
                            (public_key.e.bit_length() + 7) // 8, "big"
                        ),
                    ),
                ]
            ),
        )

    def import_certificate(
        device: PivApp,
        key_hex: str,
        certificate: bytes,
    ) -> None:
        payload = Tlv.build(
            [
                (0x5C, bytes(bytearray.fromhex(KEY_TO_CERT_OBJ_ID_MAP[key_hex]))),
                (
                    0x53,
                    Tlv.build(
                        [
                            (0x70, certificate),
                            (0x71, bytes([0])),
                        ]
                    ),
                ),
            ]
        )

        device.send_receive(0xDB, 0x3F, 0xFF, payload)

    @nk3.group()
    @click.option(
        "--experimental",
        default=False,
        is_flag=True,
        help="Allow to execute experimental features",
    )
    def piv(experimental: bool) -> None:
        """Nitrokey PIV App"""
        check_experimental_flag(experimental)
        pass

    @piv.command(help="Authenticate with the admin key.")
    @click.argument(
        "admin-key",
        type=click.STRING,
        default=default_admin_key,
    )
    def admin_auth(admin_key: str) -> None:
        try:
            admin_key_bytes = bytearray.fromhex(admin_key)
        except ValueError:
            local_critical(
                "Key is expected to be an hexadecimal string",
                support_hint=False,
            )

        device = PivApp()
        device.authenticate_admin(admin_key_bytes)
        local_print("Authenticated successfully")

    @piv.command(help="Initialize the PIV application.")
    @click.argument(
        "admin-key",
        type=click.STRING,
        default=default_admin_key,
    )
    def init(admin_key: str) -> None:
        try:
            admin_key_bytes = bytearray.fromhex(admin_key)
        except ValueError:
            local_critical(
                "Key is expected to be an hexadecimal string",
                support_hint=False,
            )

        device = PivApp()
        device.authenticate_admin(admin_key_bytes)
        guid = device.init()
        local_print("Device intialized successfully")
        local_print(f"GUID: {guid.hex().upper()}")

    @piv.command(help="Print information about the PIV application.")
    def info() -> None:
        device = PivApp()
        serial_number = device.serial()
        local_print(f"Device: {serial_number}")
        reader = device.reader()
        local_print(f"Reader: {reader}")
        guid = device.guid()
        local_print(f"GUID: {guid.hex().upper()}")

    @piv.command(help="Change the admin key.")
    @click.option(
        "--current-admin-key",
        type=click.STRING,
        default=default_admin_key,
        help="Current admin key.",
    )
    @click.argument(
        "new-admin-key",
        type=click.STRING,
    )
    def change_admin_key(current_admin_key: str, new_admin_key: str) -> None:
        try:
            current_admin_key_bytes = bytearray.fromhex(current_admin_key)
            new_admin_key_bytes = bytearray.fromhex(new_admin_key)
        except ValueError:
            local_critical(
                "Key is expected to be an hexadecimal string",
                support_hint=False,
            )

        device = PivApp()
        device.authenticate_admin(current_admin_key_bytes)
        device.set_admin_key(new_admin_key_bytes)
        local_print("Changed key successfully")

    @piv.command(help="Change the PIN.")
    @click.option(
        "--current-pin",
        type=click.STRING,
        prompt="Enter the current PIN",
        hide_input=True,
        help="Current PIN.",
    )
    @click.option(
        "--new-pin",
        type=click.STRING,
        prompt="Enter the new PIN",
        hide_input=True,
        help="New PIN.",
    )
    def change_pin(current_pin: str, new_pin: str) -> None:
        if len(new_pin) > 8 or len(new_pin) < 6 or not new_pin.isdigit():
            local_critical(
                "PIV application PIN must consist of 6 to 8 numeric characters",
                support_hint=False,
            )
        device = PivApp()
        device.change_pin(current_pin, new_pin)
        local_print("Changed pin successfully")

    @piv.command(help="Change the PUK.")
    @click.option(
        "--current-puk",
        type=click.STRING,
        prompt="Enter the current PUK",
        hide_input=True,
        help="Current PUK.",
    )
    @click.option(
        "--new-puk",
        type=click.STRING,
        prompt="Enter the new PUK",
        hide_input=True,
        help="New PUK.",
    )
    def change_puk(current_puk: str, new_puk: str) -> None:
        device = PivApp()
        device.change_puk(current_puk, new_puk)
        local_print("Changed puk successfully")

    @piv.command(help="Reset the retry counter.")
    @click.option(
        "--puk",
        type=click.STRING,
        prompt="Enter the PUK",
        hide_input=True,
        help="Current PUK.",
    )
    @click.option(
        "--new-pin",
        type=click.STRING,
        prompt="Enter the new PIN",
        hide_input=True,
        help="New PIN.",
    )
    def reset_retry_counter(puk: str, new_pin: str) -> None:
        device = PivApp()
        device.reset_retry_counter(puk, new_pin)
        local_print("Unlocked PIN successfully")

    @piv.command(help="Reset the PIV application.")
    def factory_reset() -> None:
        device = PivApp()
        try:
            device.factory_reset()
        except ValueError:
            local_critical(
                "Factory reset could not be performed. You first need to lock the PIN with 3 failed attempts",
                support_hint=False,
            )
        local_print("Factory reset successfully")

    KEY_TO_CERT_OBJ_ID_MAP = {
        "9A": "5FC105",
        "9C": "5FC10A",
        "9D": "5FC10B",
        "9E": "5FC101",
        "82": "5FC10D",
        "83": "5FC10E",
        "84": "5FC10F",
        "85": "5FC110",
        "86": "5FC111",
        "87": "5FC112",
        "88": "5FC113",
        "89": "5FC114",
        "8A": "5FC115",
        "8B": "5FC116",
        "8C": "5FC117",
        "8D": "5FC118",
        "8E": "5FC119",
        "8F": "5FC11A",
        "90": "5FC11B",
        "91": "5FC11C",
        "92": "5FC11D",
        "93": "5FC11E",
        "94": "5FC11F",
        "95": "5FC120",
    }

    def _validate_rfc4514(
        ctx: click.core.Context, param: click.core.Option, value: str
    ) -> Optional[x509.Name]:
        if value is None:
            return value

        try:
            subject_name = x509.Name.from_rfc4514_string(value)
            return subject_name
        except ValueError:
            raise click.BadParameter("Must be valid RFC4514 string.")

    @piv.command(help="Import a key and a certificate from a .p12 file")
    @click.option(
        "--admin-key",
        type=click.STRING,
        default=default_admin_key,
        help="Current admin key",
    )
    @click.option(
        "--key",
        type=key_id_click_type,
        default=default_key_id,
        help="Key slot for operation.",
    )
    @click.option(
        "--path",
        type=click.Path(allow_dash=True),
        default="-",
        help="Path to the .pem file containing the private key",
    )
    def import_key(
        admin_key: str,
        key: str,
        path: str,
    ) -> None:
        try:
            admin_key_bytes = bytearray.fromhex(admin_key)
        except ValueError:
            local_critical(
                "Key is expected to be an hexadecimal string",
                support_hint=False,
            )

        with open(path, "rb") as key_file:
            private_key, certificate, _ = pkcs12.load_key_and_certificates(
                key_file.read(), password=None
            )
        if (
            not isinstance(private_key, rsa.RSAPrivateKey)
            or private_key.key_size != 2048
            or certificate is None
        ):
            local_critical(
                "--path must point to a RSA 2048 private key and certificate as a p12 file",
                support_hint=False,
            )
            return

        key_hex = key.upper()
        key_ref = int(key_hex, 16)

        piv_app = PivApp()
        piv_app.authenticate_admin(admin_key_bytes)
        import_rsa2048(
            piv_app,
            key_ref,
            private_key.private_numbers(),
            private_key.public_key().public_numbers(),
        )

        import_certificate(piv_app, key_hex, certificate.public_bytes(Encoding.DER))

        return

    @piv.command(help="Generate a new key and certificate signing request.")
    @click.option(
        "--admin-key",
        type=click.STRING,
        default=default_admin_key,
        help="Current admin key",
    )
    @click.option(
        "--key",
        type=key_id_click_type,
        default=default_key_id,
        help="Key slot for operation.",
    )
    @click.option(
        "--algo",
        type=click.Choice(
            ["rsa2048", "rsa3072", "rsa4096", "nistp256"], case_sensitive=False
        ),
        default="nistp256",
        help="Algorithm for the key.",
    )
    @click.option(
        "--subject-name",
        type=click.STRING,
        callback=_validate_rfc4514,
        help="Subject name for the certificate signing request.",
    )
    @click.option(
        "--subject-alt-name-upn",
        type=click.STRING,
        help="Subject alternative name (UPN) for the certificate signing request.",
    )
    @click.option(
        "--pin",
        type=click.STRING,
        prompt="Enter the PIN",
        hide_input=True,
        help="Current PIN.",
    )
    @click.option(
        "--path",
        type=click.Path(allow_dash=True),
        default="-",
        help="Write certificate signing request to path.",
    )
    def generate_key(
        admin_key: str,
        key: str,
        algo: str,
        subject_name: Optional[x509.Name],
        subject_alt_name_upn: Optional[str],
        pin: str,
        path: str,
    ) -> None:
        try:
            admin_key_bytes = bytearray.fromhex(admin_key)
        except ValueError:
            local_critical(
                "Key is expected to be an hexadecimal string",
                support_hint=False,
            )
        key_hex = key.upper()
        key_ref = int(key_hex, 16)

        device = PivApp()
        device.authenticate_admin(admin_key_bytes)
        device.login(pin)

        algo = algo.lower()
        if algo == "rsa2048":
            algo_id = b"\x07"
        elif algo == "rsa3072":
            algo_id = b"\x05"
        elif algo == "rsa4096":
            algo_id = b"\x16"
        elif algo == "nistp256":
            algo_id = b"\x11"
        else:
            local_critical("Unimplemented algorithm", support_hint=False)

        if algo in ("rsa2048", "rsa3072", "rsa4096"):
            key_selector = {
                "rsa2048": b"\x10",
                "rsa3072": b"\x18",
                "rsa4096": b"\x20",
            }[algo]
        elif algo in ("nistp256",):  # TODO: add "nistp384" later
            key_selector = {
                "nistp256": b"\x01",
            }[algo]
        else:
            local_critical("Unimplemented algorithm", support_hint=False)

        body = Tlv.build([(0xAC, Tlv.build([(0x80, algo_id), (0x81, key_selector)]))])

        ins = 0x47
        p1 = 0
        p2 = key_ref
        response = device.send_receive(ins, p1, p2, body)

        data = Tlv.parse(response)
        data_tmp = find_by_id(0x7F49, data)
        if data_tmp is None:
            local_critical("Device did not send public key data")
            return

        data = Tlv.parse(data_tmp)

        if algo == "nistp256":
            key_data = find_by_id(0x86, data)
            if key_data is None:
                local_critical("Device did not send public key data")
                return
            key_data = key_data[1:]
            public_x = int.from_bytes(key_data[:32], byteorder="big", signed=False)
            public_y = int.from_bytes(key_data[32:], byteorder="big", signed=False)
            public_numbers_ecc = ec.EllipticCurvePublicNumbers(
                public_x,
                public_y,
                cryptography.hazmat.primitives.asymmetric.ec.SECP256R1(),
            )
            public_key_ecc = public_numbers_ecc.public_key()
        elif algo in ("rsa2048", "rsa3072", "rsa4096"):
            modulus_data = find_by_id(0x81, data)
            exponent_data = find_by_id(0x82, data)
            if modulus_data is None or exponent_data is None:
                local_critical("Device did not send public key data")
                return

            modulus = int.from_bytes(modulus_data, byteorder="big", signed=False)
            exponent = int.from_bytes(exponent_data, byteorder="big", signed=False)
            public_numbers_rsa = rsa.RSAPublicNumbers(exponent, modulus)
            public_key_rsa = public_numbers_rsa.public_key()
        else:
            local_critical("Unimplemented algorithm")

        certificate_builder = x509.CertificateBuilder()
        csr_builder = x509.CertificateSigningRequestBuilder()

        if subject_name is None:
            crypto_rdns = x509.Name([])
        else:
            crypto_rdns = subject_name

        certificate_builder = (
            certificate_builder.subject_name(crypto_rdns)
            .issuer_name(crypto_rdns)
            .not_valid_before(datetime.datetime(2000, 1, 1, 0, 0))
            .not_valid_after(datetime.datetime(2099, 1, 1, 0, 0))
            .serial_number(x509.random_serial_number())
        )
        csr_builder = csr_builder.subject_name(crypto_rdns)

        # SEQUENCE
        # SEQUENCE
        # OBJECT            :aes-256-cbc
        # SEQUENCE
        # OBJECT            :id-aes256-wrap
        # SEQUENCE
        # OBJECT            :aes-192-cbc
        # SEQUENCE
        # OBJECT            :id-aes192-wrap
        # SEQUENCE
        # OBJECT            :aes-128-cbc
        # SEQUENCE
        # OBJECT            :id-aes128-wrap
        # SEQUENCE
        # OBJECT            :des-ede3-cbc
        # SEQUENCE
        # OBJECT            :des-cbc
        # SEQUENCE
        # OBJECT            :rc2-cbc
        # INTEGER           :80
        # SEQUENCE
        # OBJECT            :rc4
        # INTEGER           :0200
        smime_extension = bytes(
            bytearray.fromhex(
                "308183300B060960864801650304012A300B060960864801650304012D300B0609608648016503040116300B0609608648016503040119300B0609608648016503040102300B0609608648016503040105300A06082A864886F70D0307300706052B0E030207300E06082A864886F70D030202020080300E06082A864886F70D030402020200"
            )
        )

        crypto_extensions: Sequence[Tuple[x509.ExtensionType, bool]] = [
            (x509.BasicConstraints(ca=False, path_length=None), True),
            (
                x509.KeyUsage(
                    digital_signature=True,
                    content_commitment=True,
                    key_encipherment=False,
                    data_encipherment=False,
                    key_agreement=False,
                    key_cert_sign=False,
                    crl_sign=False,
                    encipher_only=False,
                    decipher_only=False,
                ),
                True,
            ),
            (
                x509.ExtendedKeyUsage(
                    [
                        x509.oid.ExtendedKeyUsageOID.CLIENT_AUTH,
                        x509.oid.ExtendedKeyUsageOID.SMARTCARD_LOGON,
                    ]
                ),
                False,
            ),
            (
                x509.UnrecognizedExtension(
                    oid=x509.oid.ObjectIdentifier("1.2.840.113549.1.9.15"),
                    value=smime_extension,
                ),
                False,
            ),
        ]

        for ext, critical in crypto_extensions:
            certificate_builder = certificate_builder.add_extension(ext, critical)
            csr_builder = csr_builder.add_extension(ext, critical)

        if subject_alt_name_upn is not None:
            crypto_sujbect_alt_name = x509.SubjectAlternativeName(
                [
                    x509.OtherName(
                        x509.ObjectIdentifier("1.3.6.1.4.1.311.20.2.3"),
                        # bytes, because it's different from bytearray, and tlv because
                        # it expects already DER encoded ASN1
                        bytes(
                            Tlv.build([(0x0C, subject_alt_name_upn.encode("utf-8"))])
                        ),
                    )
                ]
            )
            certificate_builder = certificate_builder.add_extension(
                crypto_sujbect_alt_name, False
            )
            csr_builder = csr_builder.add_extension(crypto_sujbect_alt_name, False)

        if algo == "nistp256":
            # 9C PIN requires login to be the operation just before
            if key_ref == 0x9C:
                device.login(pin)
            csr = csr_builder.sign(
                P256PivSigner(device, key_ref, public_key_ecc), hashes.SHA256()
            )
            if key_ref == 0x9C:
                device.login(pin)
            certificate = certificate_builder.public_key(public_key_ecc).sign(
                P256PivSigner(device, key_ref, public_key_ecc), hashes.SHA256()
            )
        elif algo in ("rsa2048", "rsa3072", "rsa4096"):
            if key_ref == 0x9C:
                device.login(pin)
            csr = csr_builder.sign(
                RsaPivSigner(device, key_ref, public_key_rsa), hashes.SHA256()
            )
            if key_ref == 0x9C:
                device.login(pin)
            certificate = certificate_builder.public_key(public_key_rsa).sign(
                RsaPivSigner(device, key_ref, public_key_rsa), hashes.SHA256()
            )
        else:
            local_critical("Unimplemented algorithm")

        with click.open_file(path, mode="wb") as file:
            file.write(csr.public_bytes(Encoding.DER))

        import_certificate(device, key_hex, certificate.public_bytes(Encoding.DER))

    @piv.command(help="Write a certificate to a key slot.")
    @click.argument(
        "admin-key",
        type=click.STRING,
        default=default_admin_key,
    )
    @click.option(
        "--format",
        type=click.Choice(["DER", "PEM"], case_sensitive=False),
        default="PEM",
        help="Format of certificate.",
    )
    @click.option(
        "--key",
        type=key_id_click_type,
        default=default_key_id,
        help="Key slot for operation.",
    )
    @click.option(
        "--path",
        type=click.Path(allow_dash=True),
        default="-",
        help="Write certificate to path.",
    )
    def write_certificate(admin_key: str, format: str, key: str, path: str) -> None:
        try:
            admin_key_bytes: bytes = bytearray.fromhex(admin_key)
        except ValueError:
            local_critical(
                "Key is expected to be an hexadecimal string",
                support_hint=False,
            )

        device = PivApp()
        device.authenticate_admin(admin_key_bytes)

        with click.open_file(path, mode="rb") as f:
            cert_bytes = f.read()
        format = format.upper()
        if format == "DER":
            cert_serialized = cert_bytes
            cert = x509.load_der_x509_certificate(cert_bytes)
        elif format == "PEM":
            cert = x509.load_pem_x509_certificate(cert_bytes)
            cert_serialized = cert.public_bytes(Encoding.DER)

        payload = Tlv.build(
            [
                (0x5C, bytes(bytearray.fromhex(KEY_TO_CERT_OBJ_ID_MAP[key.upper()]))),
                (0x53, Tlv.build([(0x70, cert_serialized), (0x71, bytes([0]))])),
            ]
        )

        device.send_receive(0xDB, 0x3F, 0xFF, payload)

    @piv.command(help="Read a certificate from a key slot.")
    @click.option(
        "--format",
        type=click.Choice(["DER", "PEM"], case_sensitive=False),
        default="PEM",
        help="Format of certificate.",
    )
    @click.option(
        "--key",
        type=key_id_click_type,
        default=default_key_id,
        help="Key slot for operation.",
    )
    @click.option(
        "--path",
        type=click.Path(allow_dash=True),
        default="-",
        help="Read certificate from path.",
    )
    def read_certificate(format: str, key: str, path: str) -> None:
        device = PivApp()

        value = device.cert(
            bytes(bytearray.fromhex(KEY_TO_CERT_OBJ_ID_MAP[key.upper()]))
        )

        if value is None:
            print("Certificate not found", file=sys.stderr)
            return

        format = format.upper()
        if format == "DER":
            cert_serialized = value
            x509.load_der_x509_certificate(value)
        elif format == "PEM":
            cert = x509.load_der_x509_certificate(value)
            cert_serialized = cert.public_bytes(Encoding.PEM)

        with click.open_file(path, mode="wb") as f:
            f.write(cert_serialized)

    @piv.command(help="List certificates.")
    def list_certificates() -> None:
        device = PivApp()

        headers = ["Slot", "Algorithm", "Subject", "Serial Number", "Issuer"]
        data = []

        for key, slot in KEY_TO_CERT_OBJ_ID_MAP.items():
            cert = device.cert(bytes(bytearray.fromhex(slot)))
            if cert is not None:
                parsed_cert = x509.load_der_x509_certificate(cert)
                data.append(
                    [
                        key,
                        parsed_cert.signature_algorithm_oid._name,
                        parsed_cert.subject.rfc4514_string(),
                        f"{parsed_cert.serial_number:x}",
                        parsed_cert.issuer.rfc4514_string(),
                    ]
                )

        if data:
            print_table(headers, data)
        else:
            local_print("No certificate found.")

    @piv.command(help="Get Windows authentication certificate mapping.")
    def get_windows_auth_mapping() -> None:
        device = PivApp()

        cert = device.cert(bytes(bytearray.fromhex(KEY_TO_CERT_OBJ_ID_MAP["9A"])))
        if cert is not None:
            parsed_cert = x509.load_der_x509_certificate(cert)

            issuer_name = parsed_cert.issuer
            issuer_name.rdns.reverse()
            issuer_name_reversed = issuer_name.rfc4514_string()

            serial_number = bytearray.fromhex(f"{parsed_cert.serial_number:x}")
            serial_number.reverse()
            serial_number_reversed = serial_number.hex()

            subject_key_identifier: Union[None, str]
            try:
                subject_key_identifier = parsed_cert.extensions.get_extension_for_class(
                    x509.SubjectKeyIdentifier
                ).value.digest.hex()
            except x509.ExtensionNotFound:
                subject_key_identifier = None

            public_key = parsed_cert.public_bytes(Encoding.DER)
            digest = hashes.Hash(hashes.SHA1())
            digest.update(public_key)
            public_key_hash = digest.finalize()
            sha1_public_key = public_key_hash.hex()

            local_print(
                "Set mapping in 'altSecurityIdentities' attribute to one of the following:"
            )
            local_print(
                f"X509IssuerSerialNumber: X509:<I>{issuer_name_reversed}<SR>{serial_number_reversed}"
            )
            if subject_key_identifier:
                local_print(
                    f"               X509SKI: X509:<SKI>{subject_key_identifier}"
                )
            local_print(f"     X509SHA1PublicKey: X509:<SHA1-PUKEY>{sha1_public_key}")
        else:
            local_print("No certificate found.")

except ImportError:
    from pynitrokey.cli.nk3.pcsc_absent import PCSC_ABSENT

    @nk3.group(
        invoke_without_command=True,
        context_settings=dict(
            ignore_unknown_options=True,
        ),
    )
    @click.argument("args", nargs=-1, type=click.UNPROCESSED)
    def piv(args: list[str]) -> None:
        """Nitrokey PIV App"""
        local_critical(PCSC_ABSENT, support_hint=False)
